import collections
import contextlib

import torch
import pprint
import numpy as np
import fnmatch
from typing import Any, List, Union, Optional, Dict, Callable, Tuple
import omlet.utils as U


class DataDict(collections.abc.Mapping):
    _PROTECTED_METHODS = [
        # basic dict-like methods
        "keys",
        "values",
        "items",
        "get",
        "pop",
        "update",
        "settings",
        "Immutable",
        # functional transforms
        "to_dict",
        "to_numpy",
        "to_tensor",
        "map",
        "multi_map",
        "traverse",
        "multi_traverse",
        # type methods
        "is_array",
        "is_primitive_array",
        "is_advanced_array",  # supports advanced indexing: np.ndarray and torch.Tensor
        "is_mapping",
        "is_immutable",
        "is_sliceable",
    ]

    class Immutable:
        """
        Use this wrapper to prevent DataDict from converting the dict content into DataDict
        recursively or index into an np.ndarray/torch.Tensor. For example:

        b = DataDict()
        b.obs = {'rgb': 1, 'sensor': 2}  # b.obs will become another DataDict structure
        b.obs = DataDict.Immutable({'rgb': 1, 'sensor': 2})  # b.obs will remain a primitive dict
                    # b.obs.rgb will NOT work, you have to use b.obs['rgb']
        b.sensor = np.ndarray([1,2,3])
        b[:2]  # will index into b.sensor
        b.sensor = DataDict.Immutable(np.ndarray([1,2,3]))
        b[:2]  # will not index b.sensor and will return it as-is
        """

        def __init__(self, data):
            self.data = data

        def __repr__(self):
            return "<immu> " + repr(self.data)

    _DEFAULT_SETTINGS = {
        # b[:2] will index into primitive sequence like list and tuple
        # by default, only np.ndarray and torch.Tensor supports indexing
        "index_primitive_sequence": True,
        # if True, b[2:5] = other_b will check for keys in other_b that don't exist in b
        "strict_assign": True,
    }

    # controls global setting
    _settings = _DEFAULT_SETTINGS.copy()

    def __init__(self, _data_: collections.abc.Mapping = None, **kwargs):
        """
        Args:
            DataDict(mapping) positional only
            or DataDict(**kwargs)
        """
        self._internal_keys = []
        if _data_ is not None:
            assert not kwargs, (
                "DataDict can only be constructed by either a single "
                "positional arg (of type Mapping) or **kwargs, but not both"
            )
            if isinstance(_data_, self.__class__):
                items = _data_._wrapped_items()
            else:
                items = _data_.items()
        else:
            items = kwargs.items()

        for k, v in items:
            self[k] = v

    def to_dict(self):
        """
        Convert to primitive dict
        """
        return {
            k: v.to_dict() if isinstance(v, DataDict) else v for k, v in self.items()
        }

    @staticmethod
    def is_array(obj):
        return DataDict.is_primitive_array(obj) or DataDict.is_advanced_array(obj)

    @staticmethod
    def is_primitive_array(obj):
        return isinstance(obj, collections.abc.MutableSequence)

    @staticmethod
    def is_advanced_array(obj):
        return isinstance(obj, (np.ndarray, torch.Tensor))

    @staticmethod
    def is_mapping(obj):
        return isinstance(obj, collections.abc.Mapping)

    @classmethod
    def is_immutable(cls, obj):
        """
        Neither sequence nor mapping.
        If `index_primitive_sequence` is False, python list is also considered immutable.
        """
        return not cls.is_sliceable(obj) and not DataDict.is_mapping(obj)

    @classmethod
    def is_sliceable(cls, obj):
        if cls._settings["index_primitive_sequence"]:
            return cls.is_array(obj)
        else:
            return cls.is_advanced_array(obj)

    @staticmethod
    def _is_private(name):
        return name.startswith("_")

    @classmethod
    def settings(cls, **kwargs):
        """
        Any None will be ignored
        """
        if kwargs.pop("reset", False):
            cls._settings = cls._DEFAULT_SETTINGS.copy()
        for k, v in kwargs.items():
            assert k in cls._DEFAULT_SETTINGS, f"setting {k} is not supported."
            if v is not None:
                cls._settings[k] = v

    @classmethod
    @contextlib.contextmanager
    def settings_scope(cls, **kwargs):
        old_settings = cls._settings.copy()
        cls.settings(**kwargs)
        yield
        cls._settings = old_settings

    @staticmethod
    def _unwrap_immutable(x):
        return x.data if isinstance(x, DataDict.Immutable) else x

    def __getattribute__(self, name):
        """
        WARNING: __getattr__ only gets called every time an attribute is missing
            __getattribute__ gets called every time an attribute is *accessed*
        """
        if DataDict._is_private(name) or name in DataDict._PROTECTED_METHODS:
            return object.__getattribute__(self, name)
        else:
            try:
                return self.__getitem__(name)
            except KeyError as e:
                raise AttributeError(str(e))

    def __setattr__(self, name, value):
        if DataDict._is_private(name):
            return object.__setattr__(self, name, value)
        else:
            try:
                self.__setitem__(name, value)
            except KeyError as e:
                raise AttributeError(str(e))

    def __delattr__(self, name):
        if name in DataDict._PROTECTED_METHODS:
            raise AttributeError(f"{name} is a protected method, cannot be deleted")
        elif DataDict._is_private(name):
            raise AttributeError(f"{name} is a private attribute, cannot be deleted")
        else:
            try:
                self.__delitem__(name)
            except KeyError as e:
                raise AttributeError(str(e))

    def __getitem__(self, index: Union[str, slice]):
        if isinstance(index, str):
            if "." in index:
                parent_key, child_key = index.split(".", 1)
                # recurse through every `.`
                return self[parent_key][child_key]
            else:  # base case, key does not have "."
                if index not in self._internal_keys:
                    raise KeyError(f"Missing key {index}")
                ret = self.__dict__[index]
                return self._unwrap_immutable(ret)

        else:  # array slicing
            ret = DataDict()
            for k, v in self._wrapped_items():
                if self.is_immutable(v):
                    # primitive type cannot be index-sliced
                    ret[k] = v
                elif self.is_primitive_array(v):
                    assert self._settings["index_primitive_sequence"], (
                        "INTERNAL, should have been handled by "
                        "self._is_immutable() already"
                    )
                    # support advanced indexing into list
                    v_slice = np.array(v, dtype=np.object)[index]
                    ret[k] = type(v)(v_slice)
                else:
                    # ndarray and tensor
                    ret[k] = v[index]
            return ret

    def __setitem__(self, index: str, value):
        if isinstance(index, str):
            if "." in index:
                parent_key, child_key = index.split(".", 1)
                if parent_key not in self:
                    self[parent_key] = DataDict()
                # recurse through every `.`
                self[parent_key][child_key] = value
            else:  # base case, key does not have "."
                if index in DataDict._PROTECTED_METHODS:
                    raise KeyError(
                        f"{index} is a protected method, cannot be overridden"
                    )
                elif DataDict._is_private(index):
                    raise KeyError(f"DataDict keys must not start with '_'")
                if (
                    DataDict.is_mapping(value)
                    # we don't deepcopy item that's already DataDict
                    and not isinstance(value, DataDict)
                ):
                    self.__dict__[index] = DataDict(value)
                else:
                    self.__dict__[index] = value
                if index not in self._internal_keys:
                    self._internal_keys.append(index)

        else:  # array slicing
            assert self.is_mapping(value), (
                f"assignment {value} must be a mapping that "
                f"partially matches the structure of this DataDict object"
            )
            if not isinstance(value, DataDict):
                value = DataDict(value)
            self._check_same_struct(value, prefix="")
            for k, assign in value._wrapped_items():
                if k in self.keys():
                    ours = self[k]
                    assign = self._unwrap_immutable(assign)
                    # support advanced indexing into list
                    if self.is_primitive_array(ours):
                        ours = np.array(ours, dtype=np.object)
                        ours.__setitem__(index, assign)
                        self[k] = ours
                    else:
                        # if it's a sequence, then assign the value to the slice portion
                        # if it's another DataDict, recurse
                        self[k].__setitem__(index, assign)

    def __delitem__(self, key):
        assert isinstance(key, str), "only string key deletion is supported"
        if "." in key:
            parent_key, child_key = key.split(".", 1)
            if parent_key not in self._internal_keys:
                raise KeyError(f"Parent key {parent_key} does not exist for deletion")
            self.__dict__[parent_key].__delitem__(child_key)
        else:
            if key in self._internal_keys:
                self._internal_keys.remove(key)
            else:
                raise KeyError(f'Key "{key}" does not exist')

    def _check_same_struct(self, other_datadict, prefix: str):
        """
        Check if the assignable portions are the same struct
        Only list, tuple, np.ndarray and torch.Tensor are checked
        """
        for k, assign in other_datadict._wrapped_items():
            new_prefix = f"{prefix}.{k}".lstrip(".")
            if k not in self.keys():
                if self._settings["strict_assign"]:
                    raise KeyError(
                        f"DataDict does not have key {new_prefix}, cannot assign. Alternatively, you can set b.settings(strict_assign=False)"
                    )
                else:
                    continue

            if isinstance(assign, DataDict):
                ours = self[k]
                assert self.is_mapping(ours), (
                    f"key {k} -> value: {ours} is not a mapping, "
                    f"cannot be assigned another mapping"
                )
                ours._check_same_struct(assign, new_prefix)
            else:
                assert self.is_sliceable(
                    self[k]
                ), f"key {k} -> value: {self[k]} is not assignable."

    def keys(self):
        return self._internal_keys

    def values(self):
        return (self[k] for k in self.keys())

    def items(self):
        return ((k, self[k]) for k in self.keys())

    def _wrapped_items(self):
        # with Immutable wrap. self.items() will unwrap the Immutables
        return ((k, self.__dict__[k]) for k in self.keys())

    def get(self, key: str, default: Optional[Any] = None):
        if key in self:  # invokes __contains__ recursively
            return self[key]
        else:
            return default

    def pop(self, key: str, default: Optional[Any] = None):
        if key in self:
            v = self[key]
            del self[key]
            return v
        else:
            return default

    def update(self, other: collections.abc.Mapping):
        if not isinstance(other, DataDict):
            other = DataDict(other)
        for k, v in other.items():
            self[k] = v

    def _preprocess_dtypes(self, dtypes, mode):
        if dtypes is None:
            dtypes = {}
        elif DataDict.is_mapping(dtypes):
            assert not any(DataDict.is_mapping(dt) for dt in dtypes.values()), (
                f"dtype spec {dtypes} cannot be nested, please use dot notation "
                f"to indicated nested keys, e.g. obs.next.sensor"
            )
            dtypes = dtypes.copy()
        else:
            dtypes = {None: dtypes}

        if mode == "numpy":
            return dtypes  # numpy can accept string as type
        elif mode == "torch":
            for k, v in dtypes.items():
                # convert string "float32" to torch.float32
                if isinstance(v, str):
                    dtypes[k] = getattr(torch, v)
            return dtypes
        else:
            raise NotImplementedError("INTERNAL", mode)

    def _match_pattern(self, key, pattern_dict):
        """
        Returns:
            matched value
        """
        for pattern, v in pattern_dict.items():
            if pattern is not None and fnmatch.fnmatch(key, pattern):
                return v

        if None in pattern_dict:  # catch-all key
            return pattern_dict[None]
        else:
            return None  # use default dtype

    def to_numpy(
        self,
        inplace: bool = True,
        dtype: Union[Dict[str, Union[str, type]], str, type, None] = None,
        copy: bool = False,
        non_blocking: bool = False,
        include_keys: Optional[List[str]] = None,
        exclude_keys: Optional[List[str]] = None,
    ):
        """
        dtypes: one of None, np.dtype, or {key_name: np.dtype}
            1. None: use default dtype inferred from the data
            2. np.dtype: use this dtype for all values
            3. a dict that maps a key to desired dtype
               nested key should be specified with dots, e.g. `obs.rgb`
               special key `None` to be a "catch-all" key
               you can also use special value `None` to automatically infer dtype
               for a given key
        """
        dtype = self._preprocess_dtypes(dtype, "numpy")

        def _convert_fn(key, value):
            if not U.match_patterns(
                key, include_keys, exclude_keys, precedence="exclude"
            ):
                # key is not matched, we don't convert and return value as-is
                return value
            return U.any_to_numpy(
                value,
                dtype=self._match_pattern(key, dtype),
                copy=copy,
                non_blocking=non_blocking,
                smart_optimize=True,
            )

        return self.map(_convert_fn, inplace=inplace, with_key=True)

    def to_tensor(
        self,
        inplace: bool = True,
        dtype: Union[Dict[str, Union[str, torch.dtype]], str, torch.dtype, None] = None,
        copy: bool = False,
        device: Union[torch.device, int, str, None] = None,
        non_blocking: bool = False,
        include_keys: Optional[List[str]] = None,
        exclude_keys: Optional[List[str]] = None,
    ):
        """
        dtypes: one of None, np.dtype, or {key_name: np.dtype}
            1. None: use default dtype inferred from the data
            2. np.dtype: use this dtype for all values
            3. a dict that maps a key to desired dtype
               nested key should be specified with dots, e.g. `obs.rgb`
               special key `None` to be a "catch-all" key
               you can also use special value `None` to automatically infer dtype
               for a given key

        device: one of None, str, or torch.device
        non_blocking: transfer to device asynchronously, see torch.Tensor.to()
        copy: force copy even when tensor dtype and device don't change, see torch.Tensor.to()
        """
        dtype = self._preprocess_dtypes(dtype, "torch")

        def _convert_fn(key, value):
            if not U.match_patterns(
                key, include_keys, exclude_keys, precedence="exclude"
            ):
                # key is not matched, we don't convert and return value as-is
                return value
            return U.any_to_torch_tensor(
                value,
                dtype=self._match_pattern(key, dtype),
                device=device,
                copy=copy,
                non_blocking=non_blocking,
                smart_optimize=True,
            )

        return self.map(_convert_fn, inplace=inplace, with_key=True)

    def _map_helper(
        self,
        fn: Callable[..., Any],
        other_datadicts,
        inplace: bool,
        prefix: str,
        with_key: bool,
    ):
        """
        Args:
            fn: (k, value) -> new_value, will only be applied to non-DataDict-type leaf nodes
        """
        ret = self if inplace else DataDict()
        for k, v in self.items():
            new_prefix = f"{prefix}.{k}".lstrip(".")  # artifact from recursion
            # unwrap to support arbitrary dict type
            other_values = [DataDict._unwrap_immutable(d[k]) for d in other_datadicts]
            if isinstance(v, DataDict):
                assert all(
                    DataDict.is_mapping(other_v) for other_v in other_values
                ), f"Map failure at key {k}: other dicts do not have the same structure"
                new_v = v._map_helper(
                    fn,
                    other_values,
                    inplace=inplace,
                    prefix=new_prefix,
                    with_key=with_key,
                )
                if not inplace:
                    # we don't assign back to `self` if inplace
                    ret[k] = new_v
            else:
                if with_key:
                    ret[k] = fn(new_prefix, v, *other_values)
                else:
                    ret[k] = fn(v, *other_values)
        return ret

    def map(
        self,
        fn: Callable[..., Any],
        *other_datadicts,
        inplace: bool = False,
        with_key: bool = False,
    ):
        return self._map_helper(
            fn, other_datadicts, inplace, prefix="", with_key=with_key
        )

    @staticmethod
    def multi_map(fn: Callable[..., Any], *datadicts, with_key: bool = False):
        """
        Args:        
            fn(*values) -> new_value
        """
        assert len(datadicts) >= 1, "must have at least one DataDict"
        d0 = datadicts[0]
        if not isinstance(d0, DataDict):
            d0 = DataDict(d0)
        return d0.map(fn, *datadicts[1:], inplace=False, with_key=with_key)

    def _traverse_helper(
        self, fn: Callable[..., None], other_datadicts, prefix: str, with_key: bool
    ):
        """
        Args:
            fn: (k, value) -> new_value, will only be applied to non-DataDict-type leaf nodes
        """
        for k, v in self.items():
            new_prefix = f"{prefix}.{k}".lstrip(".")  # artifact from recursion
            # unwrap to support arbitrary dict type
            other_values = [DataDict._unwrap_immutable(d[k]) for d in other_datadicts]
            if isinstance(v, DataDict):
                assert all(
                    DataDict.is_mapping(other_v) for other_v in other_values
                ), f"Traversal failure at key {k}: other dicts do not have the same structure"
                v._traverse_helper(
                    fn, other_values, prefix=new_prefix, with_key=with_key
                )
            else:
                if with_key:
                    fn(new_prefix, v, *other_values)
                else:
                    fn(v, *other_values)

    def traverse(
        self, fn: Callable[..., None], *other_datadicts, with_key: bool = False
    ):
        return self._traverse_helper(fn, other_datadicts, prefix="", with_key=with_key)

    @staticmethod
    def multi_traverse(fn: Callable[..., None], *datadicts, with_key: bool = False):
        assert len(datadicts) >= 1, "must have at least one DataDict"
        d0 = datadicts[0]
        if not isinstance(d0, DataDict):
            d0 = DataDict(d0)
        d0.traverse(fn, *datadicts[1:], with_key=with_key)

    def __contains__(self, key: str):
        if "." in key:
            parent_key, child_key = key.split(".", 1)
            return self.__dict__[parent_key].__contains__(child_key)
        else:
            return key in self._internal_keys

    def __iter__(self):
        return iter(self.keys())

    def __len__(self) -> int:
        return len(self.keys())

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        INDENT = 4
        s = cls_name + " {\n"
        flag = False
        for k in self._internal_keys:
            rpl = "\n" + " " * INDENT
            obj = pprint.pformat(self.__dict__[k]).replace("\n", rpl)
            s += " " * INDENT + f"{k}: {obj},\n"
            flag = True
        if flag:
            s += "}"
        else:
            s = cls_name + "{}"
        return s



import logging
from omlet.utils.logging_utils import omlet_logger

# ================= Patch up PL lightning ================
import pytorch_lightning
import omlet.utils.misc_utils

pytorch_lightning.core.memory.get_human_readable_count = lambda num: omlet.utils.misc_utils.get_human_readable_count(
    num, precision=2
)

# ======================================
from .core import ExtendedModule, STAGES
from .trainer import *
from .callbacks import *
from .checkpoint import ExtendedCheckpoint
from .multi_optimizer import MultiOptimizerModule



"""
Quick hack: enhanced pytorch-lightning mixin
"""

import argparse
import collections
import functools
from typing import Union, Dict, List, Optional, Any
from pprint import pprint
import omlet.utils as U
import omlet.distributed as dist
import torch

from pytorch_lightning import LightningModule
from pytorch_lightning.trainer import Trainer
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning.utilities import rank_zero_only, rank_zero_warn
import logging
from . import omlet_logger


STAGES = ("train", "val", "test")


class ExtendedModule(LightningModule):
    """
    hparams should have the following keys:
        - train_metrics
        - val_metrics
        - test_metrics (defaults to val_metrics if unspecified)
        - best_metrics
        - batch_size or global_batch_size
        - eval_batch_size or global_eval_batch_size (defaults to `batch_size` if unspecified)
        - num_workers or global_num_workers

    Useful attributes:
        - hparams
        - use_ddp, use_dp, use_ddp2
        - global_step
        - current_epoch
        - hparams: a plain dictionary
            LightningModule can only handle dumping dict
        - conf: OmegaConf of the hyperparameters
        - C: alias for conf

    Patched methods:
        - {training|validation|test}_step()
        - {training|validation|test}_epoch_end()

    Overrideable method:
        - get_batch_size(batch): current returns batch[0].size(0)
            override if you have a more complicated batch structure

    best_metrics:
        {'val/acc1': 'max', 'test/acc5': 'max', 'val/loss': 'min'}
    """

    def __new__(cls, *args, **kwargs):
        """
        Class-level patch to avoid multiprocessing pickle error:
            _pickle.PicklingError: Can't pickle <function .. at ..>: it's not
            the same object as __main__.<obj>
        """
        obj = super().__new__(cls)
        # patch up subclass methods:
        #   - [train|validation|test]_step
        for stage in STAGES:
            cls._patch_pl_step(stage)
        return obj

    def __init__(self, hparams: Union[dict, argparse.Namespace, DictConfig]):
        super().__init__()

        if not isinstance(hparams, DictConfig):
            if isinstance(hparams, argparse.Namespace):
                hparams = vars(hparams)
            assert isinstance(
                hparams, collections.abc.Mapping
            ), "hparams must be a dict or Mapping object"
            hparams = OmegaConf.create(hparams)
        self.cfg = hparams
        self.hparams = OmegaConf.to_container(self.cfg, resolve=True)

        train_metrics: List[str] = self._check_hparams("train_metrics")
        val_metrics: List[str] = self._check_hparams("val_metrics")
        test_metrics: List[str] = self._check_hparams(
            "test_metrics", default=val_metrics
        )
        best_metrics: Dict[str, str] = self._check_hparams("best_metrics", default={})
        self.debug = self._check_hparams("debug", default=False)

        self._metrics_meter = {}
        self._train_step_meter = {}  # show fine-grained step logs for training
        self._metrics_history = []  # list of {'train/loss': .., 'val/acc1': ...}
        # best_metrics: {'val/acc1': {'value': 76.1, 'epoch': 87, 'step': 900}}
        self._best_metrics_values = {}
        self._best_metrics_spec = best_metrics

        for stage, names in zip(STAGES, (train_metrics, val_metrics, test_metrics)):
            self._metrics_meter[stage] = {m: U.AverageMeter() for m in names}
            if stage == "train":
                self._train_step_meter = {m: U.AverageMeter() for m in names}

        for name, value in best_metrics.items():
            assert "/" in name
            assert value in ["min", "max"]

        self._is_training_started = False
        self._train_log_cache = []  # for training with multiple optimizers

    def _check_hparams(self, key, default: Any = "__required__"):
        if key not in self.cfg and default == "__required__":
            raise KeyError(f"required key {key} not found in hparams")
        return self.cfg.get(key, default)

    @property
    def C(self):
        return self.cfg

    @property
    def num_training_batches(self):
        return self.trainer.num_training_batches

    @property
    def num_val_batches(self):
        return self.trainer.num_val_batches

    @property
    def num_test_batches(self):
        return self.trainer.num_test_batches

    @property
    def rank(self):
        return self.trainer.proc_rank

    @property
    def world_size(self):
        return self.trainer.world_size

    @property
    def device(self):
        return self.trainer.root_gpu

    @property
    def num_gpus(self):
        return self.trainer.num_gpus

    @property
    def batch_idx(self):
        return self.trainer.batch_idx

    @property
    def num_optimizers(self):
        return len(self.trainer.optimizers)

    # ==================== subclass overrideable ====================
    def get_batch_size(self, batch):
        """
        This function can be overridden in subclass pl_module
        if you have a complicated batch data structure
        """
        images, *_ = batch
        return int(images.size(0))

    # ==================== Data loaders ====================
    def _divide_by_gpu(self, name, default="__required__"):
        """
        check for the following configs:
        - batch_size or global_batch_size
        - eval_batch_size or global_eval_batch_size
        - num_workers or global_num_workers

        local version = global version / num_gpus
        """
        C = self.cfg
        if name in C:
            local_value = C[name]
            assert local_value > 0, f"{name} must > 0"
        elif f"global_{name}" in C:
            global_value = C[f"global_{name}"]
            assert (
                global_value % self.num_gpus == 0
            ), f"global_{name} {global_value} must divide number of GPUs {self.num_gpus}"
            local_value = global_value // self.num_gpus
        else:
            if default == "__required__":
                raise KeyError(
                    f"You must specify at least one of {name} or global_{name}"
                )
            else:
                local_value = default
        return local_value

    def get_dataloader(self, dataset, stage):
        assert stage in STAGES
        num_workers = self._divide_by_gpu("num_workers", default=8)
        if stage == "train":
            return torch.utils.data.DataLoader(
                dataset=dataset,
                batch_size=self._divide_by_gpu("batch_size"),
                shuffle=not self.use_ddp,
                num_workers=num_workers,
                pin_memory=True,
            )
        else:
            eval_batch_size = self._divide_by_gpu("eval_batch_size", default=-1)
            if eval_batch_size == -1:
                # defaults to training batch_size
                eval_batch_size = self._divide_by_gpu("batch_size")
            return torch.utils.data.DataLoader(
                dataset=dataset,
                batch_size=eval_batch_size,
                shuffle=False,
                num_workers=num_workers,
                pin_memory=True,
            )

    # ================ Patch [train|validation|test]_step() ===================
    @classmethod
    def _patch_pl_step(cls, stage):
        """
        This function does not change any logic, but only adds more OMLET internal
        variables for training_step_end() and validation_step_end(). The added variables are
          - omlet/batch_size: user can override get_batch_size()
          - omlet/optimizer_idx: if exists, otherwise None
        """
        pl_method_name = {
            "train": "training_step",
            "val": "validation_step",
            "test": "test_step",
        }[stage]
        pl_step_method = getattr(cls, pl_method_name)

        @functools.wraps(pl_step_method)
        def _wrapped(self: ExtendedModule, batch, batch_idx, *args, **kwargs):
            assert (
                not self.use_dp or self.use_ddp2
            ), "ExtendedModule does not yet support DP or DDP2, only single GPU or DDP"
            output = pl_step_method(self, batch, batch_idx, *args, **kwargs)
            output["omlet"] = {}
            output["omlet"]["batch_size"] = self.get_batch_size(batch)
            output["omlet"]["optimizer_idx"] = None
            # if more than one optimizer, add opt index
            if stage == "train" and U.func_has_arg(pl_step_method, "optimizer_idx"):
                assert len(args) > 0, "INTERNAL: should have optimizer_idx arg"
                output["omlet"]["optimizer_idx"] = int(args[0])
            return output
            # return self._after_step(stage, output)

        # monkey patch
        setattr(cls, pl_method_name, _wrapped)

    def _step_end(self, stage, output: Dict):
        """
        For training with multiple optimizers (e.g. in RL and GAN),
        we only return `log` dict at the last optimizer,
        i.e. optimizer_idx == self.num_optimizers() - 1
        The intermediate steps will return only {'loss': loss_value} without log
        """
        assert (
            "omlet" in output
        ), 'INTERNAL: step_end `output` should contain "omlet" sub-dict'
        om_vars = output.pop("omlet")
        batch_size = om_vars["batch_size"]
        self._update_epoch_metrics(stage, output, batch_size)
        if "progress_bar" not in output:
            output["progress_bar"] = {}
        output["progress_bar"].update(
            self._get_avg_epoch_metrics(stage, name_template="{stage}/{name}")
        )
        if stage == "train":
            oi = om_vars["optimizer_idx"]
            if oi is None:
                output_metrics = output
            else:  # starting with first optimizer in the training step
                if oi == 0:
                    self._train_log_cache = []  # reset at first optimizer
                self._train_log_cache.append(
                    {
                        k: v
                        for k, v in output.items()
                        if k not in ["loss", "progress_bar"]
                    }
                )
                if oi == self.num_optimizers - 1:
                    # last optimizer, consolidate cache
                    assert len(self._train_log_cache) == self.num_optimizers, "INTERNAL"
                    output_metrics = {}
                    for _cache in self._train_log_cache:
                        # comment out dup key check:
                        # dup_keys = set(_cache.keys()) & set(output_metrics.keys())
                        # if dup_keys:
                        #     self.log_warn(f'Multiple-optimizer training_step has duplicate log metric: {dup_keys}. Later `optimizer_idx` will overwrite the earlier ones.')
                        output_metrics.update(_cache)
                else:
                    # don't write any logs, just go to the next optimizer
                    return output

            # record log at every training batch step
            # for test and val, we don't record log at every batch step
            # only one summary statistic at step end
            # training stats are averaged over trainer.row_log_interval
            # (how often we write to TB logs)
            if "log" not in output:
                output["log"] = {}

            self._update_train_step_metrics(output_metrics, batch_size)
            output["log"].update(
                self._get_avg_train_step_metrics(name_template="train/stepwise_{name}")
            )
            self._add_extended_log(output["log"])
            # we don't do (self.batch_idx+1) because we need to sync with PL
            if self.batch_idx % self.trainer.row_log_interval == 0:
                self._reset_train_step_metrics()
        else:
            assert (
                om_vars["optimizer_idx"] is None
            ), "INTENRAL: optimizer_idx should always be None for validation and testing"
        return output

    # overriding *_step_end() methods
    def training_step_end(self, output):
        return self._step_end("train", output)

    def validation_step_end(self, output):
        return self._step_end("val", output)

    # tests are delegated to validation_step unless otherwise overridden
    def test_step(self, *args, **kwargs):
        return self.validation_step(*args, **kwargs)

    def test_step_end(self, *args, **kwargs):
        return self.validation_step_end(*args, **kwargs)

    def test_dataloader(self):
        return self.val_dataloader()

    # ============ Patch [train|validation|test]_epoch_end() ==================
    # @classmethod
    # def _patch_pl_epoch_end(cls, stage):
    #     "we don't need outputs because values are stored in self._metrics_sum"
    #     pl_method_name = {
    #         "train": "training_epoch_end",
    #         "val": "validation_epoch_end",
    #         "test": "test_epoch_end",
    #     }[stage]
    #     pl_method = getattr(cls, pl_method_name)
    #
    #     @functools.wraps(pl_method)
    #     def _wrapped(self, outputs):
    #         pl_method(self, outputs)
    #         return self._epoch_end(stage)
    #
    #     setattr(cls, pl_method_name, _wrapped)

    def _epoch_end(self, stage):
        """
        Collect stats from all processes at the end of an epoch
        """
        _info = self._get_avg_epoch_metrics(stage, name_template="{name}")
        # cumulative batch size on all GPUs should be exactly the same
        # so we can do a simple mean
        info_short_name = self.reduce(_info, op="mean")
        # add long name (train/acc1)
        info_long_name = {f"{stage}/{name}": v for name, v in info_short_name.items()}
        pbar = info_long_name.copy()
        log = info_long_name.copy()

        if self._is_training_started:  # avoid record in val sanity check
            if (
                len(self._metrics_history) == 0
                or self._metrics_history[-1]["epoch"] != self.current_epoch
            ):
                new_epoch_info = {
                    "epoch": self.current_epoch,
                    "train": {},
                    "val": {},
                    "test": {},
                }
                new_epoch_info[stage].update(info_short_name)
                self._metrics_history.append(new_epoch_info)
            else:
                self._metrics_history[-1][stage].update(info_short_name)

            for m, new_value in info_long_name.items():
                # m has format `val/acc1`
                if m in self._best_metrics_spec:
                    self._update_best_metrics(m, new_value)
                    # change name to `val/best_acc1` and add to log for TensorBoard
                    _stage, _name = m.split("/")
                    assert (
                        stage == _stage
                    ), f"metric {m} does not conform to stage {stage}"
                    best_name = f"{_stage}/best_{_name}"
                    # pbar[best_name] = self._get_best_str(m)  # pbar has too much info
                    log[best_name] = self._get_best_value(m)

        # self.log_debug(stage, 'history', self._metrics_history)
        # self.log_debug(stage, 'best', self._best_metrics_values)

        log["step"] = self.current_epoch  # set TB x-axis to epoch
        self._add_extended_log(log)
        return {"progress_bar": pbar, "log": log}

    def training_epoch_end(self, outputs):
        return self._epoch_end("train")

    def validation_epoch_end(self, outputs):
        return self._epoch_end("val")

    # tests are delegated to validation_epoch_end unless otherwise overridden
    def test_epoch_end(self, outputs):
        return self.validation_epoch_end(outputs)

    # ==================== distributed ====================
    def is_master(self):
        if self.use_ddp:
            assert dist.get_rank() == self.rank
        return self.rank == 0

    def is_worker(self):
        if self.use_ddp:
            assert dist.get_rank() == self.rank
        return self.rank != 0

    def reduce(
        self, values: Union[float, Dict[str, float]], op: str
    ) -> Union[float, Dict[str, float]]:
        if self.use_ddp:
            return dist.reduce_scalars(
                values, broadcast=True, device=self.device, op=op
            )
        else:
            return values

    # ==================== Metrics book-keeping ====================
    def _reset_epoch_metrics(self, stages="all"):
        for stage in STAGES if stages == "all" else [stages]:
            for meter in self._metrics_meter[stage].values():
                meter.reset()

    def _reset_train_step_metrics(self, stages="all"):
        for meter in self._train_step_meter.values():
            meter.reset()

    def _get_avg_epoch_metrics(self, stage, name_template="{name}"):
        return {
            name_template.format(stage=stage, name=name): meter.value
            for name, meter in self._metrics_meter[stage].items()
        }

    def _get_avg_train_step_metrics(self, name_template="{name}"):
        return {
            name_template.format(name=name): meter.value
            for name, meter in self._train_step_meter.items()
        }

    def _update_epoch_metrics(self, stage, output, batch_size):
        for key, value in output.items():
            if key in self._metrics_meter[stage]:
                self._metrics_meter[stage][key].update(value, batch_size)

    def _update_train_step_metrics(self, output, batch_size):
        for key, value in output.items():
            if key in self._train_step_meter:
                self._train_step_meter[key].update(value, batch_size)

    def _update_best_metrics(self, metric_name, new_value):
        assert metric_name in self._best_metrics_spec
        is_max = self._best_metrics_spec[metric_name] == "max"
        if metric_name in self._best_metrics_values:
            info = self._best_metrics_values[metric_name]
            if (
                is_max
                and float(new_value) >= info["value"]
                or not is_max
                and float(new_value) <= info["value"]
            ):
                info["value"] = new_value
                info["epoch"] = self.current_epoch
        else:
            self._best_metrics_values[metric_name] = {
                "value": new_value,
                "epoch": self.current_epoch,
            }

    def _get_best_value(self, name):
        return self._best_metrics_values[name]["value"]

    def _get_best_str(self, name):
        assert name in self._best_metrics_values
        info = self._best_metrics_values[name]
        return f'{info["value"]:.2f} @ep{info["epoch"]}'

    # ==================== Recording ====================
    def _add_extended_log(self, log_dict):
        # please use together with ExtendedTB and ExtendedWandb
        log_dict["system/global_step"] = self.global_step
        log_dict["system/epoch"] = self.current_epoch

    # ==================== Override hooks ====================
    def on_epoch_start(self):
        self._reset_epoch_metrics()
        self._reset_train_step_metrics()
        self._is_training_started = True  # avoid sanity check

    def on_save_checkpoint(self, checkpoint):
        # patch pl
        extended = {
            "metrics_history": self._metrics_history,
            "best_metrics": self._best_metrics_values,
        }
        checkpoint["extended"] = extended
        return checkpoint

    def on_load_checkpoint(self, checkpoint):
        # patch pl
        extended = checkpoint["extended"]
        self._metrics_history = extended["metrics_history"]
        self._best_metrics_values = extended["best_metrics"]

    def on_ddp_connection(self):
        """
        User override to provide logic right after multiprocessing fork
        """
        pass

    def init_ddp_connection(
        self, proc_rank: int, world_size: int, is_slurm_managing_tasks: bool = True
    ):
        """
        Override this method to add features to DDP children processes
        Callbacks can optionally define a method:
            on_ddp_connection(proc_rank, world_size)
        """
        # propagate global logging configs to children processes
        if self.C.get("omlet", {}).get("logging", None):
            U.initialize_omlet_logging(**self.C.omlet.logging)
        super().init_ddp_connection(
            proc_rank=proc_rank,
            world_size=world_size,
            is_slurm_managing_tasks=is_slurm_managing_tasks,
        )
        self.on_ddp_connection()
        for callback in self.trainer.callbacks:
            if hasattr(callback, "on_ddp_connection"):
                callback.on_ddp_connection(proc_rank, world_size)

    # ==================== debugging ====================
    @rank_zero_only  # equivalent to `if self.is_master()`
    def print(self, *args, **kwargs):
        print(*args, **kwargs)

    def _log_write(self, level, *args, **kwargs):
        if omlet_logger.isEnabledFor(level):
            if "end" not in kwargs:
                kwargs["end"] = ""
            s = U.print_str(*args, **kwargs)
            omlet_logger.log(level, s)

    @rank_zero_only
    def log_debug(self, *args, **kwargs):
        # DEBUG2 = logging.DEBUG + 2, we log higher to avoid debug messages from other apps
        self._log_write(logging.DEBUG2, *args, **kwargs)

    @rank_zero_only
    def log_info(self, *args, **kwargs):
        self._log_write(logging.INFO, *args, **kwargs)

    @rank_zero_only
    def log_infov(self, *args, **kwargs):
        # INFOV = logging.INFO - 1, more verbose
        self._log_write(logging.INFOV, *args, **kwargs)

    @rank_zero_only
    def log_infovv(self, *args, **kwargs):
        # INFOV = logging.INFO - 2, more verbose
        self._log_write(logging.INFOV, *args, **kwargs)

    @rank_zero_only
    def log_infovvv(self, *args, **kwargs):
        # INFOV = logging.INFO - 3, more verbose
        self._log_write(logging.INFOVVV, *args, **kwargs)

    @rank_zero_only
    def log_warn(self, *args, **kwargs):
        self._log_write(logging.WARNING, *args, **kwargs)

    @rank_zero_only
    def log_error(self, *args, **kwargs):
        self._log_write(logging.ERROR, *args, **kwargs)

    @rank_zero_only
    def log_critical(self, *args, **kwargs):
        self._log_write(logging.CRITICAL, *args, **kwargs)



"""
Easily configure pl.Trainer with Hydra
"""
import os
import sys
import time
from typing import Optional, Union, Dict, Any, List, Callable

import pytorch_lightning as pl
from pytorch_lightning.callbacks import Callback

from .loggers import *
from omegaconf import DictConfig, OmegaConf
import omlet.utils as U
import omlet.distributed as dist
from . import omlet_logger as _omlet_log
from .checkpoint import ExtendedCheckpoint


DEFAULT_OS_ENVS = {
    # https://discuss.pytorch.org/t/issue-with-multiprocessing-semaphore-tracking/22943/4
    "PYTHONWARNINGS": "ignore:semaphore_tracker:UserWarning",
    "OMP_NUM_THREADS": 1,
}


def _omlet_trainer_schema(cfg: DictConfig = None):
    schema = dict(
        root_dir="???",
        run_name="???",
        epochs="???",
        gpus="???",
        fp16=False,
        resume=False,  # Union[str, int, bool]
        monitor_metric="???",
        monitor_metric_mode="auto",
        save_top_k=3,
        save_epoch_interval=5,
        always_save_last=True,
        best_filename_template=None,
        eval=False,
        # ===== ExtendedModule specific fields
        batch_size=None,  # per GPU
        global_batch_size=None,  # will be used if batch_size is not specified
        eval_batch_size=None,  # per GPU
        global_eval_batch_size=None,  # will be used if eval_batch_size is not specified
        num_workers=4,  # per GPU
        global_num_workers=None,  # will be used if num_workers is not specified
        train_metrics=[],  # e.g. [loss, acc1, acc5]
        val_metrics=[],
        best_metrics={},  # e.g. {'train/loss': 'max', 'val/loss': min}
        # ===== distributed environment variables
        master_addr="localhost",
        master_port="auto",
        node_rank=0,
        distributed_backend="ddp",  # the only thing we support now
        # ===== reproducibility
        seed=None,  # use system time
        deterministic=False,  # cuDNN deterministic mode
        # ===== callbacks
        wandb={"enable": False},
        callbacks={},  # callbacks to be instantiated
        trainer={},  # extra trainer args
        # ===== added by omlet_trainer(), user should not add directly:
        exp_dir=None,  # "{root_dir}/{run_name}"
        ckpt_dir=None,  # "{root_dir}/{run_name}/ckpt"
        omlet={},  # to be filled by initialize_omlet_config() function
    )
    schema = OmegaConf.create(schema)
    if cfg is None:
        return schema
    else:
        return OmegaConf.merge(schema, cfg)


class OmletTrainer:
    def __init__(
        self,
        pl_module_cls,
        cfg: DictConfig,
        extra_callbacks: List[Callback] = [],
        run_name_generator: Optional[Callable[[DictConfig], str]] = None,
    ):
        self._pl_module_cls = pl_module_cls
        self._run_name_generator = run_name_generator
        self.cfg = _omlet_trainer_schema(cfg)
        # init_logging=False because we need to manipulate log fpath to be relative to exp_dir
        U.initialize_omlet_config(self.cfg, init_logging=False, set_os_envs=True)
        U.set_os_envs(DEFAULT_OS_ENVS)
        self._exp_dir_exists = False
        self._extra_callbacks = extra_callbacks
        self.trainer = None
        self.model = None

    @property
    def C(self):
        return self.cfg

    def configure(self):
        """
        Configures pl.Trainer
        """
        if self.model is not None or self.trainer is not None:
            return
        self._configure_run_name()
        self._configure_exp_dir()
        self._configure_logging()
        self._configure_random()
        self._configure_distributed()
        checkpoint_callback = self._configure_checkpoint()
        logger_callbacks = self._configure_logger_callbacks()
        callbacks = self._configure_generic_callbacks()
        callbacks.extend(self._extra_callbacks)
        resume_path = self._configure_resume()
        self._dump_config_yaml()

        C = self.C
        self.model = self._pl_module_cls(C)
        self.trainer = pl.Trainer(
            gpus=C.gpus,
            max_epochs=C.epochs,
            distributed_backend=C.distributed_backend,
            precision=16 if C.fp16 else 32,
            checkpoint_callback=checkpoint_callback,
            callbacks=callbacks,
            logger=logger_callbacks,
            resume_from_checkpoint=resume_path,
            **C.trainer,
        )

    def fit(self):
        try:
            self.configure()
            return self.trainer.fit(self.model)
        except KeyboardInterrupt:
            _omlet_log.critical("KeyboardInterrupt: stop training")
            sys.exit(0)

    def test(self):
        try:
            self.configure()
            return self.trainer.test(self.model)
        except KeyboardInterrupt:
            _omlet_log.critical("KeyboardInterrupt: stop testing")
            sys.exit(0)

    def run(self):
        if self.C.eval:
            return self.test()
        else:
            return self.fit()

    def _configure_run_name(self):
        C = self.cfg
        if self._run_name_generator is not None:
            C.run_name = self._run_name_generator(C)
        # strip away trailing item_sep, such as "," or "_"
        C.run_name = C.run_name.strip(" _" + C.omlet.override_name.item_sep)
        # check run_name sanity
        for special_char in "\\/$#&|\"'~!^*:<>":
            assert (
                special_char not in C.run_name
            ), f'Run name "{C.run_name}" cannot have special character {special_char}'

    def _configure_exp_dir(self):
        """
        Sets up file structure
        """
        C = self.cfg
        # set up experiment dir file structure
        C.root_dir = os.path.expanduser(C.root_dir)
        assert os.path.isdir(C.root_dir), f"root dir {C.root_dir} is not a folder."
        C.exp_dir = U.f_join(C.root_dir, C.run_name)
        self._exp_dir_exists = os.path.exists(C.exp_dir)
        if not self._exp_dir_exists:
            U.f_mkdir(C.exp_dir)
        C.ckpt_dir = U.f_join(C.exp_dir, "ckpt")
        U.f_mkdir(C.ckpt_dir)

    def _configure_logging(self):
        C = self.cfg
        if U.is_relative_path(C.omlet.logging.file_path):
            C.omlet.logging.file_path = U.f_join(C.exp_dir, C.omlet.logging.file_path)
        U.initialize_omlet_logging(**C.omlet.logging)
        _omlet_log.info(f"Full config:\n{C.pretty(resolve=True)}")
        _omlet_log.info(f"Run name: {C.run_name}")

    def _configure_random(self):
        # if seed is None, will return the actual seed based on system time
        seed = U.set_seed_everywhere(self.C.seed, deterministic=self.C.deterministic)
        self.C.seed = seed
        if seed is not None:
            _omlet_log.info(f"Random seed: {seed}")
        if self.C.deterministic:
            _omlet_log.warn(
                "You have set cudnn to deterministic mode, will impact performance!"
            )

    def _configure_distributed(self):
        C = self.C
        if C.distributed_backend == "ddp":
            # update env variables for distributed use case
            if C.master_port == "auto":
                C.master_port = dist.random_free_tcp_port()
                _omlet_log.info(
                    f"MASTER_PORT env not specified. Randomly picks a free port: {C.master_port}"
                )
            else:
                C.master_port = int(C.master_port)
            C.node_rank = int(C.node_rank)
            _envs = {
                "MASTER_ADDR": C.master_addr,
                "NODE_RANK": C.node_rank,
                "MASTER_PORT": C.master_port,
            }
            U.set_os_envs(_envs)
        else:
            raise NotImplementedError(
                "distributed_backend=ddp is the only mode supported for now"
            )

    def _configure_checkpoint(self):
        C = self.C
        assert "/" in C.monitor_metric
        if not C.monitor_metric_mode:
            C.monitor_metric_mode = "auto"
        assert C.monitor_metric_mode in ["min", "max", "auto"]

        if not C.best_filename_template:
            C.best_filename_template = "best/{epoch}-{" + C.monitor_metric + ":.2f}"

        _omlet_log.info(f"checkpoint dir: {C.ckpt_dir}")
        return ExtendedCheckpoint(
            C.ckpt_dir,
            filename_template="{epoch}",
            best_filename_template=C.best_filename_template,
            monitor_metric=C.monitor_metric,
            monitor_metric_mode=C.monitor_metric_mode,
            save_top_k=C.save_top_k,
            save_epoch_interval=C.save_epoch_interval,
            always_save_last=C.always_save_last,
        )

    def _configure_logger_callbacks(self):
        """
        Currently supported:
            - Tensorboard
            - Wandb
        """
        C = self.C
        loggers = []
        tb_logger = ExtendedTensorBoardLogger(
            U.f_join(C.exp_dir, "tb"), name="", version=""
        )
        loggers.append(tb_logger)

        if C.wandb.pop("enable"):
            _kwargs = {
                "save_dir": C.exp_dir,
                "name": C.run_name,
                "log_model": True,
                "project": os.path.basename(os.path.normpath(C.root_dir)),
            }
            _kwargs.update(C.wandb)
            wandb_logger = ExtendedWandbLogger(**_kwargs)
            loggers.append(wandb_logger)

        return loggers

    def _configure_generic_callbacks(self):
        callbacks = []
        C = self.C
        for callback_cfg in C.get("callbacks", {}).values():
            cb = U.instantiate(callback_cfg)
            # hack link reference to config
            cb.cfg = C
            cb.run_name = C.run_name
            cb.exp_dir = C.exp_dir
            callbacks.append(cb)
        return callbacks

    def _configure_resume(self) -> str:
        C = self.C
        resume = C.resume
        if resume:
            assert isinstance(resume, (int, str, bool))
            if resume is True:
                resume = U.f_join(C.ckpt_dir, "last.ckpt")
            elif isinstance(resume, int):
                resume = U.f_join(C.ckpt_dir, f"epoch={resume}.ckpt")
            elif os.path.isabs(resume):
                resume = os.path.expanduser(resume)
                if self._exp_dir_exists:
                    _omlet_log.warn(
                        f"The destination experiment dir already exists: {C.exp_dir}, make sure you do not unintentionally overwrite old checkpoints and logs"
                    )
                    # give the user a bit time to terminate
                    time.sleep(2)
            else:
                resume = U.f_join(C.ckpt_dir, resume)
            if not resume.endswith(".ckpt"):
                resume += ".ckpt"
            # check resume ckpt path must exist
            if not os.path.exists(resume):
                raise FileNotFoundError(f"Resume file {resume} does not exist")
            _omlet_log.info(f"Resuming run from checkpoint: {resume}")
        else:
            if self._exp_dir_exists:
                _omlet_log.warn(
                    f"The destination experiment dir already exists: {C.exp_dir}, but `resume` option is not set. Make sure you do not unintentionally overwrite old checkpoints and logs"
                )
                # give the user a bit time to terminate
                time.sleep(2)
            _omlet_log.info(f"Starting a new run from scratch: {C.run_name}")
            resume = None
        return resume

    def _dump_config_yaml(self):
        orig_fpath = U.f_join(self.C.exp_dir, "config.yaml")
        fpath = U.next_available_file_name(orig_fpath, suffix_template="_v{i+1}")
        U.yaml_dump(U.to_dict(self.C), fpath)
        if fpath == orig_fpath:
            _omlet_log.info(f"config saved to {fpath}")
        else:
            _omlet_log.warning(
                f"config.yaml already exists in the experiment folder, saving to another version {fpath}"
            )



import torch
import torchvision
import numpy as np
import os
import math
from PIL import Image


def generate_dataset(
    dataset,
    output_size=64,
    patch_size=8,
    digit_ratio_range=(0.4, 0.6),
    aspect_ratio_range=(1 / 1.5, 1.5),
):

    """
    Args:
        dataset: MNIST dataset
        output_size : size of output images
        patch_size: size of patches for generating noises
        digit_ratio_range: digit_height / output_size
        aspect_ratio_range: digit_width / digit_height
    """
    import cv2

    images = []
    labels = []
    bboxes = []

    # generate dataset
    num_img = len(dataset)
    for i in range(num_img):

        if (i + 1) % 5000 == 0:
            print(f"processed {i+1}/{num_img} images")

        img = np.array(dataset[i][0])
        label = dataset[i][1]

        # random stretching

        digit_ratio = (digit_ratio_range[1] - digit_ratio_range[0]) * np.random.rand(1)[
            0
        ] + digit_ratio_range[0]
        aspect_ratio = (aspect_ratio_range[1] - aspect_ratio_range[0]) * np.random.rand(
            1
        )[0] + aspect_ratio_range[0]
        h = int(output_size * digit_ratio)
        w = min(int(h * aspect_ratio), output_size)
        img_stretch = cv2.resize(img, dsize=(w, h), interpolation=cv2.INTER_CUBIC)

        # generate noise
        noise = np.zeros((output_size, output_size), dtype=np.uint8)
        num_patch = math.ceil(output_size / patch_size)
        for j in range(num_patch * num_patch):

            # randomly sample an image and crop it
            rand_idx = np.random.randint(num_img)
            crop_index = np.random.randint(img.shape[0] - patch_size)
            sample_patch = np.array(dataset[rand_idx][0])[
                crop_index : crop_index + patch_size,
                crop_index : crop_index + patch_size,
            ]

            m = j // num_patch * patch_size
            n = j % num_patch * patch_size
            if (m + patch_size) > output_size:
                sample_patch = sample_patch[: output_size - m, :]
            if (n + patch_size) > output_size:
                sample_patch = sample_patch[:, : output_size - n]
            noise[
                m : min(m + patch_size, output_size),
                n : min(n + patch_size, output_size),
            ] = sample_patch

        # composite the image with noise
        if h == output_size:
            x = 0
        else:
            x = np.random.randint(0, output_size - h)
        if w == output_size:
            y = 0
        else:
            y = np.random.randint(0, output_size - w)

        bbox = [[x, y], [x + h, y + w]]
        img_coposite = noise
        img_coposite[bbox[0][0] : bbox[1][0], bbox[0][1] : bbox[1][1]] = img_stretch

        images.append(img_coposite)
        labels.append(label)
        bboxes.append(bbox)

    return images, labels, bboxes


def _get_file_name(is_train):
    return f'noisy_mnist_{"train" if is_train else "test"}.pt'


def save_dataset(data_root, is_train, save_dir, **configs):
    data_root = os.path.expanduser(data_root)
    raw_dataset = torchvision.datasets.MNIST(
        root=data_root, train=is_train, download=False
    )
    images, labels, bboxes = generate_dataset(raw_dataset, **configs)
    D = {}
    D["images"] = np.array(images, dtype=np.uint8)
    D["labels"] = np.array(labels, dtype=np.long)
    D["bboxes"] = np.array(bboxes, dtype=np.long)
    path = os.path.join(os.path.expanduser(save_dir), _get_file_name(is_train))
    torch.save(D, path)
    print("Saved to", path)


class NoisyMNIST(torchvision.datasets.VisionDataset):
    def __init__(
        self,
        root,
        train=True,
        include_bbox=False,
        transform=None,
        target_transform=None,
    ):
        """
        Args:
            include_bbox: True to have target as (label, bbox)
                False to just have `label` Long tensor as target
        """
        super().__init__(root, transform=transform, target_transform=target_transform)
        self.train = train  # training set or test set
        self.data_file = _get_file_name(train)
        D = torch.load(os.path.join(self.root, self.data_file))
        self.images = D["images"]
        self.labels = D["labels"]
        self.bboxes = D["bboxes"]
        self.include_bbox = include_bbox

    def __getitem__(self, index):
        """
        Args:
            index (int): Index

        Returns:
            tuple: (image, target) where target is index of the target class.
        """
        img, target = self.images[index], int(self.labels[index])
        if self.include_bbox:
            target = (target, self.bboxes[index])

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img, mode="L")
        if self.transform is not None:
            img = self.transform(img)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target

    def __len__(self):
        return len(self.labels)


if __name__ == "__main__":
    data_root = "/vision/group/small"
    save_dataset(data_root, True, f"{data_root}")
    save_dataset(data_root, False, f"{data_root}")



import os
import socket
import torch
import torch.distributed as _dist
from typing import List

try:
    import apex.parallel
    _HAS_NVIDIA_APEX = True
except ImportError:
    _HAS_NVIDIA_APEX = False


def has_nvidia_apex():
    return _HAS_NVIDIA_APEX


def has_pytorch_syncbn():
    """
    PyTorch >= 1.2 supports SyncBN
    PyTorch 1.1 supports too, but there is a process_group bug:

    It didn't pass in process_group to children module recursively
    https://github.com/pytorch/pytorch/pull/19240
    """
    from pkg_resources import parse_version
    return parse_version(torch.__version__) >= parse_version('1.2.0')


def _get_group(group):
    if group is None:
        return _dist.group.WORLD
    else:
        return group


get_rank = _dist.get_rank
get_world_size = _dist.get_world_size


def is_master(group=None):
    return _dist.get_rank(_get_group(group)) == 0


def get_backend(group=None):
    return _dist.get_backend(group)


def current_device():
    return torch.cuda.current_device()


def optimize_cudnn():
    import torch.backends.cudnn as _cudnn

    _cudnn.benchmark = True
    _cudnn.deterministic = False


def get_reduce_op(op):
    OPS = {
        "sum": _dist.ReduceOp.SUM,
        "mean": _dist.ReduceOp.SUM,  # no buildin for mean, we handle it ourselves
        "product": _dist.ReduceOp.PRODUCT,
        "max": _dist.ReduceOp.MAX,
        "min": _dist.ReduceOp.MIN,
    }
    op = op.lower()
    assert op in OPS, "invalid op: " + op
    return OPS[op]


def random_free_tcp_port():
    tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    tcp.bind(("", 0))
    addr, port = tcp.getsockname()
    tcp.close()
    return port


def get_cuda_visible_devices() -> List[int]:
    """
    parse CUDA_VISIBLE_DEVICES
    """
    if 'CUDA_VISIBLE_DEVICES' not in os.environ:
        return []
    devices = os.environ['CUDA_VISIBLE_DEVICES']
    return [int(g) for g in devices.strip(' ,').split(',')]


"""
https://pytorch.org/docs/stable/distributed.html
https://pytorch.org/docs/stable/data.html#torch.utils.data.distributed.DistributedSampler
https://pytorch.org/docs/stable/distributed.html#launch-utility

Complete working example:
https://github.com/pytorch/examples/blob/master/imagenet/main.py

Wrapper around nn.DataParallel and nn.parallel.DistributedDataParallel
"""
import collections
import os
import copy
import random
import numpy as np
import torch
import torch.nn as nn
import torch.distributed as dist
import torch.multiprocessing as mp
import torch.utils.data
import omlet.utils as U
from .manager import WorkerManager
from .utils import *


__all__ = [
    "DistLauncher",
    "ReplicaDistLauncher",
    "LocalDistLauncher",
    "DPLauncher",
    "SingleGPULauncher",
]

# https://discuss.pytorch.org/t/issue-with-multiprocessing-semaphore-tracking/22943/4
os.environ["PYTHONWARNINGS"] = "ignore:semaphore_tracker:UserWarning"


class DistLauncher:
    def __init__(
        self,
        *,
        world_size,
        url="env://",
        backend="nccl",
        gpu_indices=None,
        global_rank_offset=0,
        process_group_size=0,
        process_group_lists=None,
        use_nvidia_apex=False,
        print_options=None,
    ):
        """
        Args:
            url: defaults to env://. Replace localhost with 127.0.0.1
            backend: nccl, gloo, mpi, tcp
            process_group_size: mutually exclusive with `process_group_lists`
                divide all processes into groups with the same size.
                Only useful with DDP modes
            process_group_lists: must be a list of list of process global rank,
                e.g. [[0,1,2,3], [4,7], [5,6,8]]
                mutually exclusive with `process_group_size`
                Only useful with DDP modes
            print_options: a dict of options specific to print
                all_prefix (default '=>')
                master_prefix (default None): only active with broadcast=True,
                worker_prefix (default '<{}>'): worker_prefix.format(rank)
                pretty_format (default True): use pprint.pformat
                broadcast (default False): all worker processes print to stdout
        """
        if "//" not in url:
            url = "tcp://" + url
        url = url.replace("localhost", "127.0.0.1")
        self.url = url
        self.backend = backend
        if not gpu_indices:  # take all GPUs
            num_gpus = self.get_device_count()
            assert num_gpus > 0, "no GPUs found"
            gpu_indices = list(range(num_gpus))
        else:
            assert (
                isinstance(gpu_indices, collections.abc.Sequence)
                and len(gpu_indices) > 0
            ), f'invalid gpu_indices "{gpu_indices}"'
        self.gpu_indices = list(gpu_indices)
        self.world_size = world_size
        self.global_rank_offset = global_rank_offset
        # FIXME: print options
        self.print_options = U.get_print_multiprocess_options(print_options)

        self.use_nvidia_apex = use_nvidia_apex
        if self.use_nvidia_apex:
            assert has_nvidia_apex(), "Nvidia apex is not installed"
            assert backend == "nccl", "Nvidia apex only works with NCCL backend"

        assert not (
            process_group_size and process_group_lists
        ), "process_group_lists and process_group_size are mutually exclusive"
        self.process_group_lists = None
        if process_group_lists or process_group_size:
            assert self.is_ddp(), "process groups are only used under DDP modes"
            if process_group_lists:
                assert isinstance(process_group_lists, (list, tuple))
                # disallow overlapping groups
                gset = set()
                for pg in process_group_lists:
                    assert isinstance(
                        pg, (list, tuple)
                    ), "process_group_lists must be a list of list of process global rank"
                    assert pg, "process group must not be empty"
                    assert len(pg) == len(
                        set(pg)
                    ), "one group cannot have duplicate processes"
                    overlap = gset & set(pg)
                    assert (
                        len(overlap) == 0
                    ), 'processes "{}" cannot repeat across multiple groups'.format(
                        overlap
                    )
                    gset.update(pg)
                self.process_group_lists = process_group_lists
            else:
                # expand process_group_size into process_group_lists
                # e.g. size=2 with world_size=8: [[0,1],[2,3],[4,5],[6,7]]
                assert process_group_size > 0, "process_group_size must be positive"
                assert (
                    world_size % process_group_size == 0
                ), "process_group_size must divide world_size"
                self.process_group_lists = []
                pgs = process_group_size
                for i in range(world_size // pgs):
                    self.process_group_lists.append(list(range(i * pgs, (i + 1) * pgs)))

    @property
    def mode(self):
        """
        ddp: DistributedDataParallel
        dp: DataParallel
        single: single GPU training
        """
        if self.world_size > 0:
            if self.use_nvidia_apex:
                return "ddp_apex"
            else:
                return "ddp"
        elif len(self.gpu_indices) == 1:
            return "single"
        else:
            return "dp"

    def is_ddp(self):
        return self.mode in ["ddp", "ddp_apex"]

    @property
    def num_gpus(self):
        return len(self.gpu_indices)

    def get_device_count(self):
        return torch.cuda.device_count()

    def launch(self, worker_func, *worker_args, **worker_kwargs):
        """
        TODO: make this a decorator?

        Only run this in the master process. Will block until the processes exit.
        Args:
            worker_func: callable, needs to create a DistributedManager object
            and call `worker_entry()` function
        """
        if self.is_ddp():
            mp.spawn(
                self._worker_wrapper,
                nprocs=self.num_gpus,
                args=(worker_func, worker_args, worker_kwargs),
                join=True,
            )
        else:
            # simply run in the main thread
            self._worker_wrapper(
                local_rank=0,
                worker_func=worker_func,
                worker_args=worker_args,
                worker_kwargs=worker_kwargs,
            )

    def launch_async(self, worker_func, *worker_args, **worker_kwargs):
        """
        Non-blocking version of launch()

        Returns:
            torch.multiprocessing.SpawnContext, returns control immediately
        """
        if self.is_ddp():
            nprocs = self.num_gpus
        else:
            nprocs = 1
        # join=False to be non-blocking
        return mp.spawn(
            self._worker_wrapper,
            nprocs=nprocs,
            args=(worker_func, worker_args, worker_kwargs),
            join=False,
        )

    def _worker_wrapper(self, local_rank, worker_func, worker_args, worker_kwargs):
        global_rank = self.global_rank_offset + local_rank
        group = None
        group_ids = None
        if self.is_ddp():
            dist.init_process_group(
                backend=self.backend,
                init_method=self.url,
                world_size=self.world_size,
                rank=global_rank,
            )
            # all procs must reach `new_group()` call even when they don't belong
            if self.process_group_lists:
                for gids in self.process_group_lists:
                    g = dist.new_group(gids, backend=self.backend)
                    # if multiple groups share the same process, we only assign
                    # the process to the first group in which it appears
                    if global_rank in gids and group is None:
                        group = g
                        group_ids = gids
        if group is None:
            group = WorkerManager.DEFAULT_GROUP

        # set globally accessible env variable
        # can be read by get_local_rank()
        os.environ["TORCHX_LOCAL_RANK"] = str(local_rank)
        os.environ["TORCHX_GLOBAL_RANK"] = str(global_rank)
        os.environ["TORCHX_WORLD_SIZE"] = str(self.world_size)

        manager = WorkerManager(
            local_rank=local_rank,
            global_rank=global_rank,
            world_size=self.world_size,
            gpu_indices=self.gpu_indices,
            mode=self.mode,
            backend=self.backend,
            group=group,
            group_ids=group_ids,
            print_options=self.print_options,
        )
        with manager.device_context():
            worker_func(manager, *worker_args, **worker_kwargs)


def ReplicaDistLauncher(
    *,
    node_index,
    num_nodes,
    num_gpus=-1,
    gpu_indices=None,
    url="env://",
    backend="nccl",
    process_group_size=0,
    process_group_lists=None,
    use_nvidia_apex=False,
    print_options=None,
):
    """
    ReplicaDistLauncher assumes that each node has the same number of GPUs

    TODO: use lightning-style GPU spec:
        "2": 2 GPUs
        "-1" or "all"
        "3,4,5,6"
        [3, 4, 5, 6]
    """
    if gpu_indices is None:
        if num_gpus <= 0:
            num_gpus = torch.cuda.device_count()
        assert num_gpus > 0, "no GPU found"
        gpu_indices = list(range(num_gpus))
    else:
        assert num_gpus == -1, (
            "num_gpus and gpu_indices are mutually exclusive. "
            "If gpu_indices is specified, num_gpus must be -1."
        )
        assert (
            isinstance(gpu_indices, collections.abc.Sequence) and len(gpu_indices) > 0
        ), f'invalid gpu_indices "{gpu_indices}"'
    num_gpus = len(gpu_indices)

    return DistLauncher(
        world_size=num_nodes * num_gpus,
        url=url,
        backend=backend,
        gpu_indices=gpu_indices,
        global_rank_offset=node_index * num_gpus,
        process_group_size=process_group_size,
        process_group_lists=process_group_lists,
        use_nvidia_apex=use_nvidia_apex,
        print_options=print_options,
    )


def LocalDistLauncher(
    port,
    num_gpus=-1,
    gpu_indices=None,
    process_group_size=0,
    process_group_lists=None,
    backend="nccl",
    use_nvidia_apex=False,
    print_options=None,
):
    return ReplicaDistLauncher(
        node_index=0,
        num_nodes=1,
        num_gpus=num_gpus,
        gpu_indices=gpu_indices,
        url=f"tcp://127.0.0.1:{port}",
        process_group_size=process_group_size,
        process_group_lists=process_group_lists,
        backend=backend,
        use_nvidia_apex=use_nvidia_apex,
        print_options=print_options,
    )


def DPLauncher(gpu_indices=None, backend="nccl", print_options=None):
    return DistLauncher(
        world_size=0,
        gpu_indices=gpu_indices,
        backend=backend,
        print_options=print_options,
    )


def SingleGPULauncher(gpu_index=0, print_options=None):
    return DistLauncher(
        world_size=0, gpu_indices=[gpu_index], print_options=print_options
    )



import os
import numpy as np
import torch
import random
import time
import torch.nn as nn
from copy import deepcopy
from typing import Optional, Union


def weight_init(m):
    """Custom weight init for Conv2D and Linear layers."""
    if isinstance(m, nn.Linear):
        nn.init.orthogonal_(m.weight.data)
        if hasattr(m.bias, "data"):
            m.bias.data.fill_(0.0)
    elif isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d):
        gain = nn.init.calculate_gain("relu")
        nn.init.orthogonal_(m.weight.data, gain)
        if hasattr(m.bias, "data"):
            m.bias.data.fill_(0.0)


def set_seed_everywhere(
    seed: Optional[Union[int, str]], deterministic=False
) -> Optional[int]:
    """
    seed:
        None, "time", "sys": use scrambled system time
        int < 0, "noop", "skip": do nothing
        int >= 0: set seed
    """
    if seed is None or seed in ["time", "sys"]:
        # https://stackoverflow.com/questions/27276135/python-random-system-time-seed
        t = int(time.time() * 100000)
        seed = (
            ((t & 0xFF000000) >> 24)
            + ((t & 0x00FF0000) >> 8)
            + ((t & 0x0000FF00) << 8)
            + ((t & 0x000000FF) << 24)
        )
    if seed < 0 or seed in ["noop", "skip"]:
        return None
    os.environ["PYTHONHASHSEED"] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
    if deterministic:
        import torch.backends.cudnn as cudnn

        cudnn.deterministic = True
        cudnn.benchmark = False
    return seed


class eval_mode(object):
    def __init__(self, *models):
        self.models = models

    def __enter__(self):
        self.prev_states = []
        for model in self.models:
            self.prev_states.append(model.training)
            model.train(False)

    def __exit__(self, *args):
        for model, state in zip(self.models, self.prev_states):
            model.train(state)
        return False


# ========== module operations =========
def set_requires_grad(model, requires_grad):
    for param in model.parameters():
        param.requires_grad = requires_grad


def freeze_params(model):
    set_requires_grad(model, False)


def unfreeze_params(model):
    set_requires_grad(model, True)


def clip_grad_value(model, max_value):
    with torch.no_grad():
        nn.utils.clip_grad_value_(model.parameters(), max_value)


def clip_grad_norm(model, max_norm, norm_type=2):
    """
    Returns:
        Total norm of the parameters (viewed as a single vector).
    """
    with torch.no_grad():
        return nn.utils.clip_grad_norm_(
            model.parameters(), max_norm=max_norm, norm_type=norm_type
        )


def count_parameters(model):
    return sum(x.numel() for x in model.parameters())


def get_model_device(model):
    """
    Returns:
        first model parameter's device
    """
    return next(model.parameters()).device


def clone_model(model):
    # with torch.no_grad():
    new_model = deepcopy(model).to(get_model_device(model))
    # new_model.load_state_dict(model.state_dict())
    return new_model


def update_soft_params(net, target_net, tau):
    for param, target_param in zip(net.parameters(), target_net.parameters()):
        target_param.data.copy_(tau * param.data + (1 - tau) * target_param.data)


def tie_weights(src, trg):
    # TODO deprecate this
    assert type(src) == type(trg)
    trg.weight = src.weight
    trg.bias = src.bias


def torch_normalize(tensor: torch.Tensor, mean, std, inplace=False):
    """
    Adapted from https://pytorch.org/docs/stable/_modules/torchvision/transforms/functional.html#normalize

    Normalize a tensor image with mean and standard deviation.

    .. note::
        This transform acts out of place by default, i.e., it does not mutates the input tensor.

    See :class:`~torchvision.transforms.Normalize` for more details.

    Args:
        tensor (Tensor): Tensor image of size (C, H, W) to be normalized.
        mean (sequence): Sequence of means for each channel.
        std (sequence): Sequence of standard deviations for each channel.
        inplace(bool,optional): Bool to make this operation inplace.

    Returns:
        Tensor: Normalized Tensor image.
    """
    if not torch.is_tensor(tensor):
        raise TypeError("tensor should be a torch tensor. Got {}.".format(type(tensor)))

    if not inplace:
        tensor = tensor.clone()

    dtype = tensor.dtype
    mean = torch.as_tensor(mean, dtype=dtype, device=tensor.device)
    std = torch.as_tensor(std, dtype=dtype, device=tensor.device)
    if (std == 0).any():
        raise ValueError(
            f"std evaluated to zero after conversion to {dtype}, leading to division by zero."
        )
    if mean.ndim == 1:
        mean = mean[:, None, None]
    if std.ndim == 1:
        std = std[:, None, None]
    tensor.sub_(mean).div_(std)
    return tensor



import os
import random
import re
import hydra
import importlib.resources
import sys

from omegaconf import OmegaConf, DictConfig
from omegaconf.errors import OmegaConfBaseException
from typing import Optional, List, Union
import omlet.utils as U


_CLASS_REGISTRY = {}  # for instantiation


def resource_file_path(pkg_name, fname) -> str:
    with importlib.resources.path(pkg_name, fname) as p:
        return str(p)


def _register_omlet_resolvers():
    def _random_int(spec):
        if "-" in spec:
            start, end = spec.split("-")
            start, end = int(start), int(end)
        else:
            start, end = 0, int(spec)
        return random.randint(start, end)

    OmegaConf.register_resolver("randint", _random_int)


_register_omlet_resolvers()


def _omlet_internal_config_schema(cfg: DictConfig = None):
    schema = OmegaConf.create(
        {
            "shorthand": {},
            "override_name": {
                "kv_sep": "=",
                "item_sep": ",",
                "use_shorthand": False,  # use shorthand names for override
                "include_keys": None,
                "exclude_keys": [],
            },
            "os_envs": {},
            "job": {"override_name": None, "launch_dir": None},
            "logging": {  # omlet.utils.logging_utils
                "enable": True,
                "file_path": None,
                "level": None,
                "replace_all_stream_handlers": True,
            },
            "_internal": {"is_initialized": False},
        }
    )
    if cfg is None:
        return schema
    else:
        return OmegaConf.merge(schema, cfg)


def initialize_omlet_config(
    cfg: DictConfig, init_logging: bool = True, set_os_envs: bool = True
):
    """
    See _omlet_config_schema()
    """
    assert isinstance(cfg, DictConfig)
    OmegaConf.set_struct(cfg, False)
    om = _omlet_internal_config_schema()
    for k in cfg.get("omlet", {}):
        if k not in om:
            raise KeyError(f"Unrecognized omlet config key: {k}")
    om.update(cfg.get("omlet", {}))
    cfg.omlet = om  # this is a copy operation
    om = cfg.omlet

    om._internal.is_initialized = True
    assert is_omlet_initialized(cfg), "INTERNAL"

    # process shorthands
    shortset = set()
    for original, short in om.shorthand.items():
        if short in shortset:
            raise KeyError(f"Shorthand {short} is duplicated")
        if short in cfg:
            # user has already passed the short value
            cfg[original] = cfg[short]  # create a redirection to short name
            del cfg[short]
        shortset.add(short)

    # set os environment variables
    if set_os_envs:
        U.set_os_envs(om.os_envs)

    # set override_name
    om.job.override_name = omlet_override_name(cfg)
    om.job.launch_dir = hydra_original_dir()
    # override Hydra's chdir, switch batch to launch dir
    os.chdir(om.job.launch_dir)

    if init_logging:
        U.initialize_omlet_logging(**cfg.omlet.logging)
    return cfg


def print_config(cfg: DictConfig):
    print(cfg.pretty(resolve=True))


def is_hydra_initialized():
    return hydra.utils.HydraConfig.initialized()


def is_omlet_initialized(cfg):
    try:
        return bool(cfg.omlet._internal.is_initialized)
    except OmegaConfBaseException:
        return False


def hydra_config():
    # https://github.com/facebookresearch/hydra/issues/377
    # HydraConfig() is a singleton
    if is_hydra_initialized():
        return hydra.utils.HydraConfig().cfg.hydra
    else:
        return None


def hydra_override_arg_list() -> List[str]:
    """
    Returns:
        list ["lr=0.2", "batch=64", ...]
    """
    if is_hydra_initialized():
        return hydra_config().overrides.task
    else:
        return []


def hydra_override_name():
    if is_hydra_initialized():
        return hydra_config().job.override_dirname
    else:
        return ""


def omlet_override_name(cfg: Optional[DictConfig]) -> str:
    """
    Command line overrides, e.g. "gpus=4,arch=resnet18"

    If your cfg provides a top-level dict:

    omlet:
        override_name:
            kv_sep: =
            item_sep: ,
            include_keys: ['lr', 'momentum', 'model.*']  # supports patterns
            exclude_keys: ['lr', 'momentum', 'model.*']  # supports patterns

    `include_keys` takes precedence over `exclude_keys`
    """
    assert is_hydra_initialized() and is_omlet_initialized(cfg)

    override_cfg = cfg.omlet.override_name
    # cfg.omlet.shorthand maps original -> short
    # build a reverse lookup for short -> original
    longhand = {short: original for original, short in cfg.omlet.shorthand.items()}
    args = []
    for arg in hydra_override_arg_list():
        assert "=" in arg, f'INTERNAL ERROR: arg should contain "=": {arg}'
        key, value = arg.split("=", 1)
        if key in longhand:  # key is a shorthand
            key = longhand[key]
        append = U.match_patterns(
            key,
            include=override_cfg.include_keys,
            exclude=override_cfg.exclude_keys,
            precedence="exclude",
        )
        if append:
            # key is always a longhand by now
            if override_cfg.use_shorthand and key in cfg.omlet.shorthand:
                key = cfg.omlet.shorthand[key]
            args.append((key, value))
    args.sort()
    item_sep = override_cfg.item_sep
    kv_sep = override_cfg.kv_sep
    return item_sep.join(f"{key}{kv_sep}{value}" for key, value in args)


def hydra_original_dir(*subpaths):
    return os.path.join(hydra.utils.get_original_cwd(), *subpaths)


def disable_unknown_key_check(cfg):
    OmegaConf.set_struct(cfg, False)


# ==================== Instantiation tools ====================
def register_class(name, class_type):
    assert callable(class_type)
    _CLASS_REGISTRY[name] = class_type


def get_class(path):
    """
    First try to find the class in the registry,
    if it doesn't exist, use importlib to locate it
    """
    if path in _CLASS_REGISTRY:
        return _CLASS_REGISTRY[path]
    try:
        from importlib import import_module

        module_path, _, class_name = path.rpartition(".")
        mod = import_module(module_path)
        try:
            class_type = getattr(mod, class_name)
        except AttributeError:
            raise ImportError(
                "Class {} is not in module {}".format(class_name, module_path)
            )
        return class_type
    except ValueError as e:
        print("Error initializing class " + path, file=sys.stderr)
        raise e


def _is_instantiable(cfg: DictConfig):
    return isinstance(cfg, DictConfig) and ("cls" in cfg or "class" in cfg)


def _get_params(cfg, kwargs={}):
    if "params" in cfg:
        params = cfg["params"]
    else:
        params = cfg
    for k in params:
        if OmegaConf.is_missing(params, k) and k not in kwargs:
            raise ValueError(f'Missing required keyword arg "{k}"')
    return {k: v for k, v in params.items() if k not in ["cls", "class"]}


def instantiate(cfg: DictConfig, recursive=False, *args, **kwargs):
    """
    If 'class' key exists but not 'params' key, rest of the keys will be treated as params
    """
    assert bool("cls" in cfg) != bool(
        "class" in cfg
    ), 'to instantiate from config, one and only one of "cls" or "class" key should be provided'
    assert _is_instantiable(cfg)
    if "class" in cfg:
        cls = cfg["class"]
    else:
        cls = cfg["cls"]
    params = _get_params(cfg)
    params.update(kwargs)
    if recursive:
        for key, value in params.items():
            if _is_instantiable(value):
                # args and kwargs are only for top-level
                value = instantiate(value, recursive=True)
            params[key] = value
    try:
        class_type = get_class(cls)
        return class_type(*args, **params)
    except Exception as e:
        raise RuntimeError(f"Error instantiating {cls}: {e}")


def to_dict(conf):
    if OmegaConf.is_config(conf):
        return OmegaConf.to_container(conf, resolve=True)
    elif isinstance(conf, dict):
        return conf
    else:
        raise NotImplementedError(f"Unknown dict type: {type(conf)}")





import os
import sys
import subprocess

try:
    from mpi4py import MPI
except ImportError as e:
    print("please install mpi4py")
    raise e


def mpi_fork(n, sys_argv=()):
    """
    Fork the current script with workers

    Args:
        sys_argv: tuple of args to be appended to sys.argv

    Returns:
        -1 for parent process
        <rank> for child processes (between 0 and n-1)
    """
    assert n >= 1
    if os.getenv("IN_MPI") is None:
        env = os.environ.copy()
        env.update(MKL_NUM_THREADS="1", OMP_NUM_THREADS="1", IN_MPI="1")
        subprocess.check_call(
            ["mpirun", "-np", str(n), sys.executable]
            + sys.argv
            + list(map(str, sys_argv)),
            env=env,
        )
        return -1
    else:
        return MPI.COMM_WORLD.Get_rank()


def mpi_launch(n, sys_argv=(), parent_callback=None):
    """
    Run this as the **first line** in your main script.
    It will block the parent process and only proceed with the child processes.
    Has exactly the same effect as `mpiexec -n` on the command line

    Args:
        n: n children processes, excluding the parent
        parent_callback: calls before parent exists
    """
    rank = mpi_fork(n, sys_argv=sys_argv)
    if rank < 0:
        if parent_callback is not None:
            parent_callback()
        sys.exit(0)



import os
import re
from pytorch_lightning.utilities import rank_zero_warn, rank_zero_only
from pytorch_lightning.callbacks import ModelCheckpoint
import torch
import omlet.utils as U
from typing import Optional
from . import omlet_logger as _log


class ExtendedCheckpoint(ModelCheckpoint):
    def __init__(
        self,
        save_dir,
        filename_template: str = "{epoch}",
        best_filename_template: str = "best/{epoch}",
        monitor_metric: str = "val/loss",
        monitor_metric_mode: str = "auto",
        save_top_k: int = 1,
        save_epoch_interval: int = 1,
        always_save_last: bool = True,
    ):
        """
        Periodic checkpoint save every `period`, in addition to standard `best` save
        `epoch` is always actual epoch index + 1 (number of epochs so far)

        Verbose is controlled by global logging level.
            set logging level to INFOV (i.e. INFO - 2)
        """
        self.save_dir = os.path.expanduser(save_dir)
        if not filename_template:
            filename_template = "{epoch}"
        self.ckpt_path = os.path.join(self.save_dir, filename_template)
        if not best_filename_template:
            best_filename_template = "best-{epoch}"
        self.best_ckpt_path = os.path.join(self.save_dir, best_filename_template)
        os.makedirs(self.save_dir, exist_ok=True)
        self.always_save_last = always_save_last

        super().__init__(
            filepath=self.ckpt_path,
            monitor=monitor_metric,
            verbose=False,
            save_top_k=save_top_k,
            save_weights_only=False,
            mode=monitor_metric_mode,
            period=save_epoch_interval,
            prefix="",
        )

    def format_checkpoint_name(self, epoch, metrics, ver=None, ckpt_type="best"):
        """Override

        Example::

            >>> tmpdir = os.path.dirname(__file__)
            >>> ckpt = ModelCheckpoint(os.path.join(tmpdir, '{epoch}'))
            >>> os.path.basename(ckpt.format_checkpoint_name(0, {}))
            'epoch=0.ckpt'
            >>> ckpt = ModelCheckpoint(os.path.join(tmpdir, '{epoch:03d}'))
            >>> os.path.basename(ckpt.format_checkpoint_name(5, {}))
            'epoch=005.ckpt'
            >>> ckpt = ModelCheckpoint(os.path.join(tmpdir, '{epoch}-{val_loss:.2f}'))
            >>> os.path.basename(ckpt.format_checkpoint_name(2, dict(val_loss=0.123456)))
            'epoch=2-val_loss=0.12.ckpt'
            >>> ckpt = ModelCheckpoint(os.path.join(tmpdir, '{missing:d}'))
            >>> os.path.basename(ckpt.format_checkpoint_name(0, {}))
            'missing=0.ckpt'
        """
        if ckpt_type == "best":
            fpath = self.best_ckpt_path
        else:
            fpath = self.ckpt_path
        # check if user passed in keys to the string
        groups = re.findall(r"(\{.*?)[:\}]", fpath)
        assert (
            len(groups) > 0
        ), f"INTERNAL ERROR: file name template {fpath} should have at least one {{...}}"

        metrics = {k.replace("/", "_"): v for k, v in metrics.items()}
        metrics["epoch"] = epoch + 1

        for tmp in groups:
            name = tmp[1:]
            name = name.replace("/", "_")
            fpath = fpath.replace(tmp, name + "={" + name)
            if name not in metrics:
                metrics[name] = 0
        fpath = fpath.format(**metrics)
        if fpath.endswith(".ckpt"):
            fpath = fpath[: -len(".ckpt")]
        str_ver = f"_v{ver}" if ver is not None else ""
        return fpath + str_ver + ".ckpt"

    def check_monitor_top_k(self, current):
        """
        Override warning from super class
        """
        less_than_k_models = len(self.best_k_models) < self.save_top_k
        if less_than_k_models:
            return True

        if not isinstance(current, torch.Tensor):
            current = torch.tensor(current)

        monitor_op = {"min": torch.lt, "max": torch.gt,}[self.mode]

        return monitor_op(current, self.best_k_models[self.kth_best_model])

    @rank_zero_only
    def on_validation_end(self, trainer, pl_module):
        """
        Mostly copied from pytorch_lightning.callbacks.ModelCheckpoint
        """
        # only run on main process
        if trainer.proc_rank != 0:
            return

        metrics = trainer.callback_metrics
        epoch = trainer.current_epoch
        _best_path = self._save_best(metrics, epoch)
        _periodic_path = self._save_periodic(metrics, epoch, _best_path)
        self._save_last(_periodic_path, _best_path)

    def _save_best(self, metrics, epoch):
        """
        Returns:
            filepath if the best is saved, None if nothing saved
        """
        if self.save_top_k <= 0:
            # no best saved
            return None

        filepath = self.format_checkpoint_name(epoch, metrics, ckpt_type="best")
        version_cnt = 1
        while os.path.isfile(filepath):
            filepath = self.format_checkpoint_name(
                epoch, metrics, ver=version_cnt, ckpt_type="best"
            )
            # this epoch called before
            version_cnt += 1

        current = metrics.get(self.monitor)

        if current is None:
            rank_zero_warn(
                f"Can save best model only with {self.monitor} available, skipping.",
                RuntimeWarning,
            )
        elif self.check_monitor_top_k(current):
            self._do_check_save(filepath, current, epoch)
            if version_cnt > 1:
                _log.warn(
                    f"best ckpt filepath exists, saving to a different version: {filepath}"
                )
            _log.infov(
                f"\nEpoch {epoch:05d}: {self.monitor} reached"
                f" {current:0.5f} (best {self.best:0.5f}), saving model to"
                f" {filepath} as top {self.save_top_k}"
            )
            return filepath

        _log.infov(
            f"\nEpoch {epoch:03d}: {self.monitor} was not in top {self.save_top_k}"
        )
        return None

    def _save_periodic(self, metrics, epoch, _best_save_path):
        """
        Don't duplicate save if the same ckpt has already been saved as a best model
        """
        if self.period <= 0 or not ((epoch + 1) % self.period) == 0:
            # skip if not this period
            return None

        filepath = self.format_checkpoint_name(epoch, metrics, ckpt_type="periodic")
        version_cnt = 1
        while os.path.isfile(filepath):
            filepath = self.format_checkpoint_name(
                epoch, metrics, ver=version_cnt, ckpt_type="periodic"
            )
            # this epoch called before
            version_cnt += 1
        if version_cnt > 1:
            _log.warn(
                f"ckpt filepath exists, saving to a different version: {filepath}"
            )

        _log.infov(f"\nEpoch {epoch:03d}: periodic saving model to {filepath}")

        if _best_save_path:
            # the destination shouldn't exist tho
            _log.debug2(
                f"\nEpoch {epoch:03d}: periodic saving copies from the best ckpt {_best_save_path}"
            )
            U.f_copy(_best_save_path, filepath, exists_ok=True)
        else:
            self._save_model(filepath)
        return filepath

    def _save_last(self, _periodic_save_path, _best_save_path):
        if not self.always_save_last:
            return

        # always save a copy to the special `last.ckpt`
        last_path = os.path.join(self.save_dir, "last.ckpt")
        if _periodic_save_path:
            U.f_copy(_periodic_save_path, last_path, exists_ok=True)
        elif _best_save_path:
            U.f_copy(_best_save_path, last_path, exists_ok=True)
        else:
            # neither best or periodic saved
            self._save_model(last_path)



import torch
import torch.nn as nn
import torch.nn.functional as F


def mlp(input_dim, hidden_dim, output_dim, hidden_depth, output_mod=None):
    if hidden_depth == 0:
        mods = [nn.Linear(input_dim, output_dim)]
    else:
        mods = [nn.Linear(input_dim, hidden_dim), nn.ReLU(inplace=True)]
        for i in range(hidden_depth - 1):
            mods += [nn.Linear(hidden_dim, hidden_dim), nn.ReLU(inplace=True)]
        mods.append(nn.Linear(hidden_dim, output_dim))
    if output_mod is not None:
        mods.append(output_mod)
    trunk = nn.Sequential(*mods)
    return trunk


class Swish(nn.Module):
    def __init__(self, learnable_beta=False):
        super().__init__()
        self.learnable_beta = learnable_beta
        if learnable_beta:
            self.beta = nn.Parameter(torch.ones(()), requires_grad=True)
        else:
            self.beta = None

    def forward(self, x):
        if self.learnable_beta:
            return x * torch.sigmoid(self.beta * x)
        else:
            return x * torch.sigmoid(x)


class Conv2dWS(nn.Conv2d):
    """
    Weight Standardization
    """

    def forward(self, x):
        weight = self.weight
        weight_mean = (
            weight.mean(dim=1, keepdim=True)
            .mean(dim=2, keepdim=True)
            .mean(dim=3, keepdim=True)
        )
        weight = weight - weight_mean
        std = weight.view(weight.size(0), -1).std(dim=1).view(-1, 1, 1, 1) + 1e-5
        weight = weight / std.expand_as(weight)
        return F.conv2d(
            x, weight, self.bias, self.stride, self.padding, self.dilation, self.groups
        )



from typing import Union

import numpy as np
import torch


def np_dtype_size(dtype: Union[str, np.dtype]) -> int:
    return np.dtype(dtype).itemsize


_TORCH_DTYPE_TABLE = {
    torch.bool: 1,
    torch.int8: 1,
    torch.uint8: 1,
    torch.int16: 2,
    torch.short: 2,
    torch.int32: 4,
    torch.int: 4,
    torch.int64: 8,
    torch.long: 8,
    torch.float16: 2,
    torch.half: 2,
    torch.float32: 4,
    torch.float: 4,
    torch.float64: 8,
    torch.double: 8,
}


def torch_dtype(dtype: Union[str, torch.dtype, None]) -> torch.dtype:
    if dtype is None:
        return None
    elif isinstance(dtype, torch.dtype):
        return dtype
    elif isinstance(dtype, str):
        try:
            dtype = getattr(torch, dtype)
        except AttributeError:
            raise ValueError(f'"{dtype}" is not a valid torch dtype')
        assert isinstance(
            dtype, torch.dtype
        ), f"dtype {dtype} is not a valid torch tensor type"
        return dtype
    else:
        raise NotImplementedError(f"{dtype} not supported")


def torch_device(device: Union[str, int, None]) -> torch.device:
    if device is None:
        return None
    elif isinstance(device, int) and device < 0:
        return torch.device("cpu")
    else:
        return torch.device(device)


def torch_dtype_size(dtype: Union[str, torch.dtype]) -> int:
    return _TORCH_DTYPE_TABLE[torch_dtype(dtype)]


def _convert_then_transfer(x, dtype, device, copy, non_blocking):
    x = x.to(dtype=dtype, copy=copy, non_blocking=non_blocking)
    return x.to(device=device, copy=False, non_blocking=non_blocking)


def _transfer_then_convert(x, dtype, device, copy, non_blocking):
    x = x.to(device=device, copy=copy, non_blocking=non_blocking)
    return x.to(dtype=dtype, copy=False, non_blocking=non_blocking)


def any_to_torch_tensor(
    x,
    dtype: Union[str, torch.dtype, None] = None,
    device: Union[str, int, torch.device, None] = None,
    copy=False,
    non_blocking=False,
    smart_optimize: bool = True,
):
    dtype = torch_dtype(dtype)
    device = torch_device(device)

    if not isinstance(x, (torch.Tensor, np.ndarray)):
        # x is a primitive python sequence
        x = torch.tensor(x, dtype=dtype)
        copy = False

    # This step does not create any copy.
    # If x is a numpy array, simply wraps it in Tensor. If it's already a Tensor, do nothing.
    x = torch.as_tensor(x)
    # avoid passing None to .to(), PyTorch 1.4 bug
    dtype = dtype or x.dtype
    device = device or x.device

    if not smart_optimize:
        # do a single stage type conversion and transfer
        return x.to(dtype=dtype, device=device, copy=copy, non_blocking=non_blocking)

    # we have two choices: (1) convert dtype and then transfer to GPU
    # (2) transfer to GPU and then convert dtype
    # because CPU-to-GPU memory transfer is the bottleneck, we will reduce it as
    # much as possible by sending the smaller dtype

    src_dtype_size = torch_dtype_size(x.dtype)

    # destination dtype size
    if dtype is None:
        dest_dtype_size = src_dtype_size
    else:
        dest_dtype_size = torch_dtype_size(dtype)

    if x.dtype != dtype or x.device != device:
        # a copy will always be performed, no need to force copy again
        copy = False

    if src_dtype_size > dest_dtype_size:
        # better to do conversion on one device (e.g. CPU) and then transfer to another
        return _convert_then_transfer(x, dtype, device, copy, non_blocking)
    elif src_dtype_size == dest_dtype_size:
        # when equal, we prefer to do the conversion on whichever device that's GPU
        if x.device.type == "cuda":
            return _convert_then_transfer(x, dtype, device, copy, non_blocking)
        else:
            return _transfer_then_convert(x, dtype, device, copy, non_blocking)
    else:
        # better to transfer data across device first, and then do conversion
        return _transfer_then_convert(x, dtype, device, copy, non_blocking)


def any_to_numpy(
    x,
    dtype: Union[str, np.dtype, None],
    copy: bool = False,
    non_blocking: bool = False,
    smart_optimize: bool = True,
):
    if isinstance(x, torch.Tensor):
        x = any_to_torch_tensor(
            x,
            dtype=dtype,
            device="cpu",
            copy=copy,
            non_blocking=non_blocking,
            smart_optimize=smart_optimize,
        )
        return x.numpy()
    else:
        # primitive python sequence or ndarray
        return np.array(x, dtype=dtype, copy=copy)



import math
import torch.utils.data
import torch.distributed as dist


__all__ = [
    "EpochRandomSampler",
    "DistributedRandomSampler",
    "DistributedSequentialSampler",
]


class EpochRandomSampler(torch.utils.data.Sampler):
    """
    Single process sampler. Deterministic sampling based on epoch number.
    """

    def __init__(self, dataset, start_epoch, base_seed=0):
        super().__init__(dataset)
        self.dataset = dataset
        self.epoch = start_epoch
        self.base_seed = base_seed

    @property
    def seed(self):
        return self.base_seed + self.epoch

    def __iter__(self):
        # deterministically shuffle based on epoch
        g = torch.Generator()
        g.manual_seed(self.seed)
        indices = torch.randperm(len(self.dataset), generator=g).tolist()
        self.epoch += 1
        return iter(indices)

    def __len__(self):
        return len(self.dataset)


class DistributedRandomSampler(torch.utils.data.Sampler):
    """
    Similar to torch.utils.data.distributed.DistributedSampler, but can automatically
    increment epoch.
    """

    def __init__(self, dataset, start_epoch, base_seed=0, num_replicas=None, rank=None):
        """
        Args:
            base_seed: actual seed used will be `epoch + seed`
        """
        super().__init__(dataset)
        if num_replicas is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            num_replicas = dist.get_world_size()
        if rank is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            rank = dist.get_rank()
        self.dataset = dataset
        self.num_replicas = num_replicas
        self.rank = rank
        self.epoch = start_epoch
        self.base_seed = base_seed
        self.num_samples = int(math.ceil(len(self.dataset) * 1.0 / self.num_replicas))
        self.total_size = self.num_samples * self.num_replicas

    @property
    def seed(self):
        return self.base_seed + self.epoch

    def __iter__(self):
        # deterministically shuffle based on epoch
        g = torch.Generator()
        g.manual_seed(self.seed)
        indices = torch.randperm(len(self.dataset), generator=g).tolist()
        self.epoch += 1

        # add extra samples to make it evenly divisible
        indices += indices[: (self.total_size - len(indices))]
        assert len(indices) == self.total_size

        # subsample
        indices = indices[self.rank : self.total_size : self.num_replicas]
        assert len(indices) == self.num_samples

        return iter(indices)

    def __len__(self):
        return self.num_samples


class DistributedSequentialSampler(torch.utils.data.Sampler):
    """
    Sequential sampling for distributed val mode
    """

    def __init__(self, dataset, num_replicas=None, rank=None):
        super().__init__(dataset)
        if num_replicas is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            num_replicas = dist.get_world_size()
        if rank is None:
            if not dist.is_available():
                raise RuntimeError("Requires distributed package to be available")
            rank = dist.get_rank()
        self.dataset = dataset
        self.num_replicas = num_replicas
        self.rank = rank
        self.num_samples = int(math.ceil(len(self.dataset) * 1.0 / self.num_replicas))
        self.total_size = self.num_samples * self.num_replicas

    def __iter__(self):
        indices = list(range(len(self.dataset)))
        # add extra samples to make it evenly divisible
        indices += indices[: (self.total_size - len(indices))]
        assert len(indices) == self.total_size

        # subsample
        indices = indices[self.rank : self.total_size : self.num_replicas]
        assert len(indices) == self.num_samples

        return iter(indices)

    def __len__(self):
        return self.num_samples



import torch
import time
import random
import numpy as np
import itertools
from .timer import Timer
from collections import OrderedDict
from tabulate import tabulate


__all__ = ["TestCase", "BenchmarkRunner"]


class TestCase:
    def __init__(
        self,
        name,
        tags=None,
        is_cuda=False,
        is_warmup=False,
        torch_enable_grad=False,
        n_best=10,
        n_total=100,
        mean_time_unit="auto",
        std_time_unit="auto",
    ):
        """
        Args:
            is_warmup: True to run the test but ignore the output.
                Useful for CUDA tests that require warming up GPU.
        """
        self.name = name
        if tags is None:
            tags = ()
        elif isinstance(tags, str):
            tags = (tags,)
        else:
            assert isinstance(tags, (list, tuple))
            tags = tuple(tags)
        self.tags = tags
        self._is_cuda = is_cuda
        self._is_warmup = is_warmup
        if torch_enable_grad:
            self._grad_context = torch.enable_grad()
        else:
            self._grad_context = torch.no_grad()
        if n_best is None:
            n_best = n_total
        assert n_best <= n_total
        self._n_best = n_best
        self._n_total = n_total
        self._mean_time_unit = mean_time_unit
        self._std_time_unit = std_time_unit
        self._avg_best = self._avg_total = self._min = self._max = None

    def setup(self):
        pass

    def body(self):
        raise NotImplementedError

    def teardown(self):
        pass

    def _run(self, verbose):
        repeats = []
        with self._grad_context:
            self.setup()
        for _ in range(self._n_total):
            start = time.time()
            with self._grad_context:
                self.body()
            if self._is_cuda:
                torch.cuda.synchronize()
            repeats.append(time.time() - start)
            torch.cuda.empty_cache()
        self.teardown()
        repeats = np.sort(np.array(repeats))
        self._avg_best = "{} \u00B1 {}".format(
            Timer.pformat(repeats[: self._n_best].mean(), unit=self._mean_time_unit),
            Timer.pformat(repeats[: self._n_best].std(), unit=self._std_time_unit),
        )
        self._avg_total = "{} \u00B1 {}".format(
            Timer.pformat(repeats.mean(), unit=self._mean_time_unit),
            Timer.pformat(repeats.std(), unit=self._std_time_unit),
        )
        self._min = Timer.pformat(repeats.min(), unit=self._mean_time_unit)
        self._max = Timer.pformat(repeats.max(), unit=self._mean_time_unit)
        if verbose:
            print(
                "avg_best={}, avg_total={}, min={}, max={}\t{}".format(
                    self._avg_best, self._avg_total, self._min, self._max, self.name
                )
            )


class BenchmarkRunner:
    def __init__(
        self,
        test_cases,
        randomize=True,
        group_tag_cartesian=None,
        group_tag_flat=None,
        tag_headers=None,
        tag_display_mode="column",
        verbose=True,
    ):
        """
        Args:
            group_tag_cartesian: provide a list of list of tags, and generates
                a cartesian product list of tag groups
                e.g. [['forward', 'backward'], ['fp16', 'fp32']]
                generates [['forward', 'fp16'], ['forward', 'fp32'],
                        .. ['backward', 'fp16'], ['backward', 'fp32']]
                --OR--
                set group_tag_cartesian='auto' to automatically generate
                    cartesian product from your tags
            group_tag_flat: provide a list of tag lists as flat tag groups
                e.g. [['forward', 'fp16'], ['backward', 'fp32']]
                this option is mutually exclusive with group_tag_cartesian
            tag_display_mode: 'column' or 'concat'
            tag_headers: only useful if display_tag_mode=='column'
        """
        assert tag_display_mode in ["concat", "column"]
        self.warmup_tests = [c for c in test_cases if c._is_warmup]
        self.normal_tests = [c for c in test_cases if not c._is_warmup]
        assert self.warmup_tests or self.normal_tests, "must have at least one test"
        if randomize:
            random.shuffle(self.normal_tests)

        if tag_headers is None:
            tag_headers = []
        assert isinstance(tag_headers, (list, tuple))
        self._tag_headers = tag_headers

        assert group_tag_flat or group_tag_cartesian, "mutually exclusive"
        self._check_list_of_list(group_tag_flat)
        self._check_list_of_list(group_tag_cartesian)
        self._group_tag_flat = group_tag_flat
        self._tag_groups = []
        if group_tag_flat:
            self._tag_groups = group_tag_flat
            assert all(
                len(v) == len(self._tag_groups[0]) for v in self._tag_groups
            ), "all tag groups must have the same number of tags"
        self._group_tag_cartesian = group_tag_cartesian
        self._tag_groups = []  # filled by either group_tag_flat or .._cartesian
        self._tag_display_mode = tag_display_mode
        self._table = []
        self._verbose = verbose
        self.run()

    def _check_list_of_list(self, value):
        if value is not None and value != "auto":
            assert isinstance(value, (list, tuple))
            assert all(isinstance(v, (list, tuple)) for v in value)

    def _run_case(self, case):
        assert isinstance(case, TestCase)
        case._run(self._verbose)
        if not case._is_warmup:
            self._table.append(
                (
                    case.name,
                    case.tags,
                    (case._avg_best, case._avg_total, case._min, case._max),
                )
            )

    def _generate_cartesian(self):
        if self._group_tag_cartesian:
            if self._group_tag_cartesian == "auto":
                all_tags = [entry[1] for entry in self._table]
                choices = list(map(tuple, map(U.OrderedSet, zip(*all_tags))))
            else:
                choices = self._group_tag_cartesian
            self._tag_groups = list(itertools.product(*choices))

    def run(self):
        for case in self.warmup_tests + self.normal_tests:
            self._run_case(case)

        if self._verbose:
            print("*" * 40, "\n")

        self._generate_cartesian()

        # all tags are tuples
        table_groups = OrderedDict()
        for tags in self._tag_groups:
            table_groups[tuple(tags)] = []
        table_groups[None] = []  # catch the rest

        while self._table:
            name, tags, stats = self._table.pop(0)
            entry = []
            name_and_tag = [name, *tags]
            if self._tag_display_mode == "concat":
                entry.append("_".join(name_and_tag))
            elif self._tag_display_mode == "column":
                entry.extend(name_and_tag)
            entry.extend(stats)
            if tags in table_groups:
                table_groups[tags].append(entry)
            else:
                table_groups[None].append(entry)
            # if the entry's tags are a subset of one of table_group's tag group
            # for tag_group in table_groups:
            #     if tag_group and set(tag_group) in set(tags):
            #         table_groups[tag_group].append(entry)
            #         break
            # else:
            #     table_groups[None].append(entry)

        table_rows = []
        for tags, entries in table_groups.items():
            if not entries:  # keyword corresponds to nothing
                continue
            entries.sort()
            table_rows.extend(entries)
            # separator
            table_rows.append("__SEP__")
        table_rows.pop()
        # make separator same width as table
        max_widths = [0] * len(entry)
        for entry in table_rows:
            if entry == "__SEP__":
                continue
            for i, e in enumerate(entry):
                max_widths[i] = max(max_widths[i], len(e))
        for i, entry in enumerate(table_rows):
            if entry == "__SEP__":
                table_rows[i] = ("-" * width for width in max_widths)

        tag_headers = self._tag_headers[: len(entry) - 5]

        print(
            tabulate(
                table_rows,
                headers=[
                    "Run",
                    *tag_headers,
                    "avg (BEST)",
                    "avg (TOTAL)",
                    "min",
                    "max",
                ],
                tablefmt="presto",
            )
        )



import time


class Timer:
    def __init__(self):
        self._start = None
        self._end = None

    def start(self):
        self._start = time.time()

    def elapsed(self):
        self._end = time.time()
        self._interval = self._end - self._start
        return self._interval

    @staticmethod
    def pformat(interval, unit="auto", show_hms=True, float_fmt=".2f"):
        if not float_fmt:
            float_fmt = ".2f"
        if not float_fmt.startswith("{"):
            float_fmt = "{:" + float_fmt + "}"
        if unit == "auto":
            kwargs = dict(show_hms=show_hms, float_fmt=float_fmt)
            if interval >= 0.1:
                return Timer.pformat(interval, unit="s", **kwargs)
            elif interval >= 1e-4:
                return Timer.pformat(interval, unit="ms", **kwargs)
            elif interval >= 1e-7:
                return Timer.pformat(interval, unit="us", **kwargs)
            else:
                return Timer.pformat(interval, unit="ns", **kwargs)
        elif unit == "s":
            if interval >= 60 and show_hms:
                hour = int(interval // 3600)
                minute = int((interval - hour * 3600) // 60)
                sec = int(interval - 3600 * hour - 60 * minute)
                if hour == 0:
                    hour = ""
                else:
                    hour = "{}h:".format(hour)
                return "{}{:0>2d}m:{:0>2d}s".format(hour, minute, sec)
            else:
                return (float_fmt + "s").format(interval)
        elif unit == "ms":
            return (float_fmt + "ms").format(interval * 1e3)
        elif unit == "us":
            return (float_fmt + "\u03BCs").format(interval * 1e6)
        elif unit == "ns":
            return (float_fmt + "ns").format(interval * 1e9)
        else:
            raise ValueError("Unsupported unit: " + unit)

    def elapsed_str(self, unit="auto"):
        return self.pformat(self.elapsed(), unit=unit)

    def print_elapsed(self, msg="", unit="auto"):
        # in seconds:
        print(msg if msg else "elapsed", self.elapsed_str(unit=unit))

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.print_elapsed()



import os
import numpy as np
import fnmatch
import collections.abc
from typing import Optional, Dict, Any, List, Union


def get_human_readable_count(number: int, precision: int = 2):
    assert number >= 0
    labels = [" ", "K", "M", "B", "T"]
    num_digits = int(np.floor(np.log10(number)) + 1 if number > 0 else 1)
    num_groups = int(np.ceil(num_digits / 3))
    num_groups = min(num_groups, len(labels))  # don't abbreviate beyond trillions
    shift = -3 * (num_groups - 1)
    number = number * (10 ** shift)
    index = num_groups - 1
    rem = number - int(number)
    if precision > 0 and rem > 0.01:
        fmt = f"{{:.{precision}f}}"
        rem_str = fmt.format(rem).lstrip("0")
    else:
        rem_str = ""
    return f"{int(number):,d}{rem_str} {labels[index]}"


def set_os_envs(envs: Optional[Dict[str, Any]] = None):
    """
    Special value __delete__ indicates that the ENV_VAR should be removed
    """
    if envs is None:
        envs = {}
    # check for special key __delete__
    for k, v in envs.items():
        if v == "__delete__":
            os.environ.pop(k, None)
    os.environ.update({k: str(v) for k, v in envs.items() if v != "__delete__"})


def _match_patterns_helper(element, patterns):
    for p in patterns:
        if fnmatch.fnmatch(element, p):
            return True
    return False


def match_patterns(
    element,
    include: Union[str, List[str], None] = None,
    exclude: Union[str, List[str], None] = None,
    *,
    precedence: str = "exclude",
):
    """
    Args:
        include: None to disable `include` filter and delegate to exclude
        precedence: "include" or "exclude"
    """
    assert precedence in ["include", "exclude"]
    if exclude is None:
        exclude = []
    if isinstance(exclude, str):
        exclude = [exclude]
    if isinstance(include, str):
        include = [include]
    if include is None:
        # exclude is the sole veto vote
        return not _match_patterns_helper(element, exclude)

    if precedence == "include":
        return _match_patterns_helper(element, include)
    else:
        if _match_patterns_helper(element, exclude):
            return False
        else:
            return _match_patterns_helper(element, include)



import io
import os
import sys
import shlex


class PrintRedirection(object):
    """
    Context manager: temporarily redirects stdout and stderr
    """

    def __init__(self, stdout=None, stderr=None):
        """
        Args:
          stdout: if None, defaults to sys.stdout, unchanged
          stderr: if None, defaults to sys.stderr, unchanged
        """
        if stdout is None:
            stdout = sys.stdout
        if stderr is None:
            stderr = sys.stderr
        self._stdout, self._stderr = stdout, stderr

    def __enter__(self):
        self._old_out, self._old_err = sys.stdout, sys.stderr
        self._old_out.flush()
        self._old_err.flush()
        sys.stdout, sys.stderr = self._stdout, self._stderr
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.flush()
        # restore the normal stdout and stderr
        sys.stdout, sys.stderr = self._old_out, self._old_err

    def flush(self):
        "Manually flush the replaced stdout/stderr buffers."
        self._stdout.flush()
        self._stderr.flush()


class PrintToFile(PrintRedirection):
    """
    Print to file and save/close the handle at the end.
    """

    def __init__(self, out_file=None, err_file=None):
        """
        Args:
          out_file: file path
          err_file: file path. If the same as out_file, print both stdout
              and stderr to one file in order.
        """
        self.out_file, self.err_file = out_file, err_file
        if out_file:
            out_file = os.path.expanduser(out_file)
            self.out_file = open(out_file, "w")
        if err_file:
            err_file = os.path.expanduser(out_file)
            if err_file == out_file:  # redirect both stdout/err to one file
                self.err_file = self.out_file
            else:
                self.err_file = open(os.path.expanduser(out_file), "w")
        super().__init__(stdout=self.out_file, stderr=self.err_file)

    def __exit__(self, *args):
        super().__exit__(*args)
        if self.out_file:
            self.out_file.close()
        if self.err_file:
            self.err_file.close()


def PrintSuppress(no_out=True, no_err=False):
    """
    Args:
      no_out: stdout writes to sys.devnull
      no_err: stderr writes to sys.devnull
    """
    out_file = os.devnull if no_out else None
    err_file = os.devnull if no_err else None
    return PrintToFile(out_file=out_file, err_file=err_file)


class PrintString(PrintRedirection):
    """
    Redirect stdout and stderr to strings.
    """

    def __init__(self):
        self.out_stream = io.StringIO()
        self.err_stream = io.StringIO()
        super().__init__(stdout=self.out_stream, stderr=self.err_stream)

    def stdout(self):
        "Returns: stdout as one string."
        return self.out_stream.getvalue()

    def stderr(self):
        "Returns: stderr as one string."
        return self.err_stream.getvalue()

    def stdout_by_line(self):
        "Returns: a list of stdout line by line, ignore trailing blanks"
        return self.stdout().rstrip().split("\n")

    def stderr_by_line(self):
        "Returns: a list of stderr line by line, ignore trailing blanks"
        return self.stderr().rstrip().split("\n")


def print_str(*args, **kwargs):
    """
    Same as print() signature but returns a string
    """
    sstream = io.StringIO()
    kwargs.pop("file", None)
    print(*args, **kwargs, file=sstream)
    return sstream.getvalue()


def format_str(_formatter, **kwargs):
    """
    Simulate python f-string but without `f`
    """
    locals().update(kwargs)
    return eval("f" + shlex.quote(_formatter))



"""
Utils to handle nested data structures

Install dm_tree first:
https://tree.readthedocs.io/en/latest/api.html
"""
import collections

try:
    import tree

    """
    # prefix all methods with `tree_`, can be generated as:
    
    for k in tree.__all__:
        if k.isupper():
            print(f'TREE_{k} = tree.{k}')
        else:
            print(f'tree_{k} = tree.{k}')
    """
    tree_is_nested = tree.is_nested
    tree_assert_same_structure = tree.assert_same_structure
    tree_unflatten_as = tree.unflatten_as
    tree_flatten = tree.flatten
    tree_flatten_up_to = tree.flatten_up_to
    tree_flatten_with_path = tree.flatten_with_path
    tree_flatten_with_path_up_to = tree.flatten_with_path_up_to
    tree_map_structure = tree.map_structure
    tree_map_structure_up_to = tree.map_structure_up_to
    tree_map_structure_with_path = tree.map_structure_with_path
    tree_map_structure_with_path_up_to = tree.map_structure_with_path_up_to
    tree_traverse = tree.traverse

except ImportError:
    raise ImportError("Please install dm_tree first: `pip install dm_tree`")


def is_sequence(obj):
    """
    Returns:
      True if the sequence is a collections.Sequence and not a string.
    """
    return isinstance(obj, collections.abc.Sequence) and not isinstance(obj, str)


def is_mapping(obj):
    """
    Returns:
      True if the sequence is a collections.Mapping
    """
    return isinstance(obj, collections.abc.Mapping)



"""
200 classes, each 500 training images, 100 validation images
"""
import os
import random
from tqdm import tqdm


def create_mini_imagenet(
    imagenet_folder,
    dest_folder,
    num_classes: int,
    training_images_per_class: int,
    seed: int = 1,
    dry_run: bool = False,
):
    """
    Assume that `imagenet_folder` has train/ and val/ subdirs
    """
    if dry_run:
        run = print
    else:
        run = os.system

    random.seed(seed)
    imagenet_folder = os.path.expanduser(imagenet_folder)
    dest_folder = os.path.expanduser(dest_folder)
    # subsample classes
    all_classes = [
        d for d in os.listdir(f"{imagenet_folder}/train") if d.startswith("n")
    ]
    print("ImageNet classes:", len(all_classes))

    classes = random.sample(all_classes, num_classes)

    for cls in tqdm(classes):
        imgs = [
            im
            for im in os.listdir(f"{imagenet_folder}/train/{cls}")
            if im.endswith(".JPEG")
        ]
        imgs = random.sample(imgs, training_images_per_class)
        run(f"mkdir -p {dest_folder}/train/{cls}")
        for im in imgs:
            run(f"cp {imagenet_folder}/train/{cls}/{im} {dest_folder}/train/{cls}/{im}")

    run(f"mkdir -p {dest_folder}/val/")
    for cls in classes:
        run(f"cp -rf {imagenet_folder}/val/{cls} {dest_folder}/val")



"""
https://pytorch.org/docs/stable/distributed.html
https://pytorch.org/docs/stable/data.html#torch.utils.data.distributed.DistributedSampler
https://pytorch.org/docs/stable/distributed.html#launch-utility

Complete working example:
https://github.com/pytorch/examples/blob/master/imagenet/main.py

Wrapper around nn.DataParallel and nn.parallel.DistributedDataParallel
"""
import os
import copy
import random
import numpy as np
import torch
import torch.nn as nn
import torch.distributed as dist
import torch.utils.data
import omlet.utils as U
from omlet.data.loaders import *
from .utils import *


__all__ = [
    "WorkerManager",
    "get_local_rank",
    "is_master",
    "get_global_rank",
    "get_world_size",
]

# https://discuss.pytorch.org/t/issue-with-multiprocessing-semaphore-tracking/22943/4
os.environ["PYTHONWARNINGS"] = "ignore:semaphore_tracker:UserWarning"


# FIXME: merge these with pytorch get_rank, get_world_size


def get_local_rank():
    """
    We can obtain process local_rank without using WorkerManager
    """
    return int(os.environ.get("TORCHX_LOCAL_RANK", 0))


def is_master(local=True):
    if local:
        return get_local_rank() == 0
    else:
        return get_global_rank() == 0


def get_global_rank():
    return int(os.environ.get("TORCHX_GLOBAL_RANK", 0))


def get_world_size():
    return int(os.environ.get("TORCHX_WORLD_SIZE", 0))


class _AutoDeviceDataLoader(torch.utils.data.DataLoader):
    def _set_device(self, mode, device):
        self.__mode = mode
        self.__device = device

    def __iter__(self):
        for batch_data in super().__iter__():
            if not isinstance(batch_data, (tuple, list)) or len(batch_data) != 2:
                # yield without change if not in the format of (input, target) pair
                yield batch_data
            else:
                # For DataParallel, we put `input` on CPU because we don't
                # want to send the entire unsharded input to the first GPU.
                input, target = batch_data
                if self.__mode != "dp":
                    input = input.cuda(self.__device, non_blocking=True)
                target = target.cuda(self.__device, non_blocking=True)
                yield input, target


class WorkerManager:
    disable_warning = False
    DEFAULT_GROUP = dist.group.WORLD

    def __init__(
        self,
        local_rank,
        global_rank,
        world_size,
        gpu_indices,
        mode,
        backend,
        group,
        group_ids,
        print_options,
    ):
        self.local_rank = local_rank
        self.global_rank = global_rank
        self.world_size = world_size
        assert self.local_rank < len(gpu_indices), "internal error"
        self.gpu_indices = gpu_indices
        self.mode = mode
        self.backend = backend
        self.group = group  # process group
        self.group_ids = group_ids
        self.print_options = print_options
        self.print_options["process_rank"] = self.local_rank
        if "ddp" not in self.mode:
            self.print_options["broadcast"] = False

    @property
    def gpu_idx(self):
        mode = self.mode
        if mode in ["ddp", "ddp_apex", "dp"]:
            return self.gpu_indices[self.local_rank]
        elif mode == "single":
            assert len(self.gpu_indices) == 1
            return self.gpu_indices[0]
        else:
            raise NotImplementedError

    def device(self):
        return torch.device(self.gpu_idx)

    def device_context(self):
        """
        Context manager
        """
        return torch.cuda.device(self.device())

    def is_default_group(self):
        return self.group == self.DEFAULT_GROUP

    def is_ddp(self):
        return self.mode in ["ddp", "ddp_apex"]

    def is_single(self):
        return self.mode == "single"

    def is_master(self, local=True):
        if local:
            return self.local_rank == 0
        else:
            return self.global_rank == 0

    def is_worker(self):
        return not self.is_master()

    @property
    def num_gpus(self):
        return len(self.gpu_indices)

    def get_local_batch_size(self, global_batch_size):
        divisor = self.num_gpus if self.is_ddp() else 1
        return int(global_batch_size / divisor)

    def data_parallel(self, model, sync_bn=False, sync_bn_group=None, **kwargs):
        """
        Args:
            sync_bn_group: if None, defaults to syncBN over the entire `world`
        """
        mode = self.mode
        if sync_bn:
            if mode == "ddp":
                assert (
                    has_pytorch_syncbn()
                ), "must use pytorch version >= 1.2 for SyncBN. Otherwise you can use Nvidia apex syncBN."
            else:
                assert (
                    mode == "ddp_apex"
                ), "SyncBN is only supported with Nvidia Apex or pytorch version >= 1.2"
            if sync_bn_group is None:
                sync_bn_group = self.DEFAULT_GROUP

        if mode == "ddp":
            device = self.device()
            model.cuda(device)
            if sync_bn:
                model = nn.SyncBatchNorm.convert_sync_batchnorm(model, sync_bn_group)
            return nn.parallel.DistributedDataParallel(
                model, device_ids=[device], output_device=device, **kwargs
            )
        elif mode == "ddp_apex":
            assert has_nvidia_apex()
            from apex.parallel import (
                DistributedDataParallel as ApexDistributedDataParallel,
            )
            from apex.parallel import convert_syncbn_model as apex_convert_syncbn_model

            if sync_bn:
                model = apex_convert_syncbn_model(model, process_group=sync_bn_group)
            model.cuda(self.device())
            kwargs = copy.deepcopy(kwargs)
            # defaults to True unless explicitly specified
            delay_allreduce = kwargs.pop("delay_allreduce", True)
            return ApexDistributedDataParallel(
                model, delay_allreduce=delay_allreduce, **kwargs
            )
        elif mode == "dp":
            return nn.DataParallel(model, device_ids=self.gpu_indices, **kwargs).cuda()
        elif mode == "single":
            return model.cuda(self.device())
        else:
            raise NotImplementedError

    def unwrap(self, model):
        """
        Retrieve the model wrapped in DataParallel
        """
        if self.mode == "single":
            return model
        else:
            assert hasattr(
                model, "module"
            ), "internal error: dp or ddp-wrapped model has no `.module` attribute"
            return model.module

    def get_data_loader(
        self,
        dataset,
        global_batch_size,
        num_workers,
        stage,
        epoch=None,
        keep_on_cpu=False,
        **loader_kwargs
    ):
        assert stage in ["train", "val", "test"]
        mode = self.mode
        batch_size = self.get_local_batch_size(global_batch_size)
        divisor = self.num_gpus if self.is_ddp() else 1
        num_workers = int(num_workers / divisor)
        LoaderCls = (
            torch.utils.data.DataLoader if keep_on_cpu else _AutoDeviceDataLoader
        )
        if self.is_ddp():
            if stage == "train":
                assert isinstance(
                    epoch, int
                ), "`epoch` must be specified in order to use DistributedSampler in training, because it is used as the internal random seed."
                sampler = DistributedRandomSampler(dataset, start_epoch=epoch)
            else:
                sampler = DistributedSequentialSampler(dataset)
            # sampler = torch.utils.data.distributed.DistributedSampler(dataset)
            self._warning(
                """
                DistributedSampler makes the dataloader 1/num_gpu shorter. 
                Loss and accuracy values will be calculated from only 1/num_gpu data if you don't call manager.all_reduce() or .master_reduce() manually. 
                """
            )
        else:
            sampler = None
        if stage == "train":
            shuffle = sampler is None
        else:
            shuffle = False

        loader = LoaderCls(
            dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            pin_memory=True,
            sampler=sampler,
            **loader_kwargs
        )
        if isinstance(loader, _AutoDeviceDataLoader):
            loader._set_device(mode, self.device())
        return loader

    def set_seed(self, seed=111, same_worker_seed=True, deterministic=False):
        if not same_worker_seed and seed is not None:
            seed = seed + self.local_rank
        U.set_seed_everywhere(seed, deterministic=deterministic)

    def _get_reduce_op(self, op):
        # TODO fix
        OPS = {
            "sum": dist.ReduceOp.SUM,
            "mean": dist.ReduceOp.SUM,  # no buildin for mean, we handle it ourselves
            "product": dist.ReduceOp.PRODUCT,
            "max": dist.ReduceOp.MAX,
            "min": dist.ReduceOp.MIN,
        }
        op = op.lower()
        assert op in OPS, "invalid op: " + op
        return OPS[op]

    def all_reduce_(self, tensor, op="sum", group=None):
        """
        Warnings:
            input tensor will be mutated
        """
        if group is None:
            group = self.DEFAULT_GROUP
        if self.is_ddp():
            assert torch.is_tensor(tensor)
            dist.all_reduce(tensor, op=self._get_reduce_op(op), group=group)
            if op.lower() == "mean":
                tensor.div_(1.0 * dist.get_world_size())

    def all_reduce(self, tensor, op="sum", group=None):
        """
        Immutable version of all_reduce_
        Leave input tensor unchanged, return a copied reduced version
        """
        if group is None:
            group = self.DEFAULT_GROUP
        if self.is_ddp():
            assert torch.is_tensor(tensor)
            new_tensor = tensor.clone().detach()
            dist.all_reduce(new_tensor, op=self._get_reduce_op(op), group=group)
            if op.lower() == "mean":
                new_tensor.div_(1.0 * dist.get_world_size())
            return new_tensor
        else:
            return tensor

    def master_reduce_(self, tensor, op="sum", group=None):
        """
        Reduced result is only available on master proc.
        Other worker procs will leave input tensor unchanged
        Warnings:
            input tensor will be mutated on the master proc
        """
        if group is None:
            group = self.DEFAULT_GROUP
        if self.is_ddp():
            assert torch.is_tensor(tensor)
            dist.reduce(tensor, dst=0, op=self._get_reduce_op(op), group=group)
            if op.lower() == "mean" and self.is_master():
                tensor.div_(1.0 * dist.get_world_size())

    def master_reduce(self, tensor, op="sum", group=None):
        """
        Immutable version of master_reduce_
        Leave input tensor unchanged, return a copied reduced version
        """
        if group is None:
            group = self.DEFAULT_GROUP
        if self.is_ddp():
            assert torch.is_tensor(tensor)
            new_tensor = tensor.clone().detach()
            dist.reduce(new_tensor, dst=0, op=self._get_reduce_op(op), group=group)
            if op.lower() == "mean" and self.is_master():
                new_tensor.div_(1.0 * dist.get_world_size())
            return new_tensor
        else:
            return tensor

    def sync_barrier(self):
        if self.is_ddp():
            dist.barrier()

    def reduce_scalar(
        self, scalar, op="sum", *, broadcast=True, storage=None, group=None
    ):
        """
        Args:
            scalar: single python float
            broadcast: True to use all_reduce, else master_reduce
            storage: if you call this method repeatedly, consider allocating
                a persistent CUDA tensor in advance to provide temporary storage
                if None, allocates new CUDA memory every time
        Returns:
            python float
        """
        scalar = float(scalar)
        if not self.is_ddp():
            return scalar
        if storage is None:
            s = torch.tensor([scalar], dtype=torch.float32)
            if self.backend == "nccl":
                s = s.cuda(self.device())
        else:
            assert torch.is_tensor(storage)
            assert storage.numel() > 0, "storage must have at least 1 element"
            if self.backend == "nccl":
                assert storage.is_cuda
            storage[0] = scalar
            s = storage
        reduce_ = self.all_reduce_ if broadcast else self.master_reduce_
        reduce_(s, op=op, group=group)
        return float(s[0])

    def reduce_scalars(
        self, scalars, op="sum", *, broadcast=True, storage=None, group=None
    ):
        """
        Args:
            scalars: list, tuple or dict of python floats
            broadcast: True to use all_reduce, else master_reduce
            storage: if you call this method repeatedly, consider allocating
                a persistent CUDA tensor in advance to provide temporary storage
                if None, allocates new CUDA memory every time
        Returns:
            list, tuple or dict of reduced floats
        """
        if not isinstance(scalars, (tuple, list, dict)):
            # singleton falls back to reduce_scalar()
            return self.reduce_scalar(
                scalars, op=op, broadcast=broadcast, storage=storage
            )
        if not self.is_ddp():
            return scalars
        idx_map = None
        if isinstance(scalars, dict):
            values = []
            idx_map = {}
            for i, (k, v) in enumerate(scalars.items()):
                idx_map[k] = i
                values.append(float(v))
        else:
            values = [float(v) for v in scalars]
        numel = len(values)

        values = torch.tensor(values, dtype=torch.float32)
        if storage is None:
            if self.backend == "nccl":
                values = values.cuda(self.device())
        else:
            assert torch.is_tensor(storage)
            assert storage.numel() >= numel, "storage not enough"
            if self.backend == "nccl":
                assert storage.is_cuda
            storage[:numel] = values
            values = storage
        reduce_ = self.all_reduce_ if broadcast else self.master_reduce_
        reduce_(values, op=op, group=group)

        if isinstance(scalars, dict):
            return {k: float(values[idx_map[k]]) for k in idx_map}
        else:
            return type(scalars)(float(v) for v in values[:numel])

    def print(self, *objs, broadcast=None, **print_kwargs):
        """
        Args:
            broadcast: if None, inherits the broadcast print option from DistLauncher
        """
        opts = self.print_options.copy()
        if broadcast is not None:
            opts["broadcast"] = broadcast

        U.print_multiprocess(*objs, **opts, **print_kwargs)

    def printfmt(self, fmt_str, *fmt_args, broadcast=None, **fmt_kwargs):
        opts = self.print_options.copy()
        if broadcast is not None:
            opts["broadcast"] = broadcast

        U.printfmt_multiprocess(fmt_str, *fmt_args, **opts, **fmt_kwargs)



"""
Debugging purposes
"""
import os
import torch
import torch.nn as nn
import numpy as np
import pickle
from .termcolor import colored


class Tracer:
    def __init__(
        self,
        trace_file: str,
        mode: str,
        verbose: bool = True,
        atol: float = 1e-6,
        rtol: float = 1e-4,
    ):
        assert mode in ["save", "check", "check-raise", "off"]
        self.mode = mode
        self.verbose = verbose
        if self.disabled:
            return
        self.trace_file = os.path.expanduser(trace_file)
        self.trace = []
        self._pointer = 0
        self.tol = (atol, rtol)
        if self.is_check():
            assert os.path.exists(
                self.trace_file
            ), f"in check mode, trace_file must exist: {self.trace_file}"
            with open(self.trace_file, "rb") as fp:
                self.trace = pickle.load(fp)
        if self.enabled:
            self._info("TRACER INIT MODE: " + mode.upper())

    @property
    def disabled(self):
        return self.mode == "off"

    @property
    def enabled(self):
        return self.mode != "off"

    def is_check(self):
        return self.mode.startswith("check")

    def _to_float(self, value):
        if isinstance(value, nn.Module):
            return [float(m.abs().mean().item()) for m in value.parameters()]
        elif torch.is_tensor(value):
            return float(value.abs().mean().item())
        elif isinstance(value, np.ndarray):
            return float(np.mean(np.abs(value)))
        else:
            return float(value)

    def _to_shape(self, value):
        if isinstance(value, nn.Module):
            return [tuple(m.size()) for m in value.parameters()]
        elif torch.is_tensor(value):
            return tuple(value.size())
        elif isinstance(value, np.ndarray):
            return tuple(value.shape)
        else:
            return tuple()

    def _info(self, title, *args):
        ctext = colored(f"[{title}]", color="green", attrs=["bold"])
        print(ctext, *args)

    def _error(self, title, *args):
        ctext = colored(f"[{title}]", color="white", on_color="red", attrs=["bold"])
        print(ctext, *args)

    def _value_to_str(self, value):
        if isinstance(value, list):
            s = ", ".join([f"{v:.4g}" for v in value])
            return f"[{s}]"
        else:
            return f"{value:.4g}"

    def _check_diff(self, name, v, tr_v):
        atol, rtol = self.tol
        progress = (self._pointer + 1) / len(self.trace) * 100
        if abs(v - tr_v) > atol + rtol * abs(tr_v):
            self._error(
                f"VALUE DIFF {progress:>4.1f}%", f"{name}: actual {v} != saved {tr_v}"
            )
            if self.mode == "check-raise":
                raise ValueError("TRACE FAILED")
        else:
            if self.verbose:
                self._info(f"TRACE PASS {progress:>4.1f}%", name)

    def _save(self, name, value):
        if self.disabled:
            return
        assert self.mode == "save"
        shape = self._to_shape(value)
        value = self._to_float(value)
        self.trace.append((name, shape, value))
        if self.verbose:
            self._info(
                "SAVE TRACE", name, "shape=", shape, "value=", self._value_to_str(value)
            )

    def _check(self, name, value):
        tr_name, tr_shape, tr_value = self.trace[self._pointer]
        shape = self._to_shape(value)
        assert name == tr_name, f"event name does not match: {name} != {tr_name}"
        if name != tr_name:
            self._error("NAME MISMATCH", f'actual "{name}" != saved "{tr_name}"')
            raise ValueError
        if shape != tr_shape:
            self._error("SHAPE MISMATCH", f"{name} actual {shape} != saved {tr_shape}")
            raise ValueError

        value = self._to_float(value)
        if isinstance(value, list):
            ## nn.Module
            for i, (v, tr_v) in enumerate(zip(value, tr_value)):
                self._check_diff(f"Module {name}[{i}]", v, tr_v)
        else:
            self._check_diff(f"Var {name}", value, tr_value)
        self._pointer += 1

    def event(self, name, value):
        if self.disabled:
            return
        if self.mode == "save":
            self._save(name, value)
        elif self.is_check():
            self._check(name, value)
        else:
            raise NotImplementedError(self.mode)

    def done(self):
        if self.disabled:
            return
        if self.mode == "save":
            with open(self.trace_file, "wb") as fp:
                pickle.dump(self.trace, fp)
            self._info("DONE", f"debug trace saved to {self.trace_file}")
        else:
            self._info("DONE", f"trace check completed")
